#! /usr/bin/env python
# __init__.py (template)

"""This module provides an interface to menu-generation templates.

libtovid.templates should provide an assortment of "plug-in" style templates
for menu generation, and perhaps one day also for disc layout or video encoding profiles).
"""

__all__ = [\
    'textmenu',
    'thumbmenu',
    'Style']

class Style:
    """Contains style attributes, including font, color, and alignment."""
    def __init__(self, font='helvetica', fontsize=12, textcolor='white',
                 highlightcolor='green', selectcolor='red', align='center'):
        self.font = font
        self.fontsize = fontsize
        self.textcolor = textcolor
        self.highlightcolor = highlightcolor
        self.selectcolor = selectcolor
        self.align = align

"""todisc options


Style control

-showcase IMAGE|VIDEO
-textmenu
-background IMAGE|VIDEO
-menu-font
-menu-fontsize
-menu-fade
-thumb-shape
-3dthumbs
-thumb-font
-thumb-fontsize
-submenu-stroke-color
-submenu-title-color
-button-style rect|text|text-rect
-rotate
-wave
-showcase-framestyle none|glass
-title-color
-stroke-color
-highlightcolor
-selectcolor
-text-mist
-text-mist-color
-text-mist-opacity
-opacity
-blur
-thumb-mist-color
-thumb-text-color
-showcase-titles-align
-tile3x1

Audio:

-bgaudio
-menu-audiolength
-menu-audio-fade
-submenu-audio
-submenu-audiolength
-submenu-audio-fade


Other

-files
-titles
-submenus
-ani-submenus
-static
-submneu-titles
-chapters
-menu-title
-menu-length
-intro VIDEO
-seek
-showcase-seek
-loop
-playall
-chain-videos
-subtitles lang1 lang2 ...
-audio-lang chan1 chan2 ...
-aspect
-widescreen

Execution control

-debug
-noask
-tovidopts
"""



#! /usr/bin/env python
# util.py

import threading, wx
from libtovid.gui.constants import id_dict

__all__ = [\
    '_',
    'ID_to_text',
    'text_to_ID',
    'VER_GetFirstChild',
    'VideoStatSeeker']

# Hack until gettext works
def _(str):
    return str

def ID_to_text(idtype, id):
    """Convert widget ID numbers to string representations.
    """
    return id_dict[ idtype ][ id ]

def text_to_ID(txt):
    """Convert internal text identifiers into widget ID numbers
    """
    for idtype, subdict in id_dict.iteritems():
        for id, value in subdict.iteritems():
            if value == txt:
                return id
    print "text_to_ID: Couldn't match '%s'. Returning 0." % txt
    return 0

# wx.TreeCtrl.GetFirstChild workaround
def VER_GetFirstChild(obj, item):
    # For wx.Widgets >=2.5, cookie isn't needed
    if wx.MAJOR_VERSION == 2 and wx.MINOR_VERSION >= 5:
        return obj.GetFirstChild(item)
    # For other versions, use a dummy cookie
    else:
        return obj.GetFirstChild(item, 1)

class VideoStatSeeker(threading.Thread):
    """Video statistics-seeking thread class. Runs in the background determining
    duration and size of input videos."""
    # List of VideoOptions objects to gather statistics on
    listVideoOptions = []

    def __init__(self, vidOptions):
        threading.Thread.__init__(self)
        # Add vidOptions to end of queue; thread will later
        # gather statistics and store them back in vidOptions
        self.listVideoOptions.append(vidOptions)
        self.doneWithStats = False

    def run(self):
        """Run in background to gather and save statistics."""
        # For each video in queue, get and save stats
        while len(self.listVideoOptions) > 0:
            curOpts = self.listVideoOptions.pop(0)
            #curStatCmd = os.popen("idvid -terse \"%s\" | grep DURATION | sed -e \"s/DURATION://g\"" % curOpts.inFile, 'r')
            #curOpts.duration = curStatCmd.readline().strip('\n ')
            #curStatCmd.readlines()
            #curStatCmd.close()
            #print "VideoStatSeeker: duration for %s is %s" % (curOpts.inFile, curOpts.duration)

        # Done with current batch of statistics
        self.doneWithStats = True



#! /usr/bin/env python
# __init__.py (metagui)

"""A simplified GUI-writing module.


INTRODUCTION

This module demonstrates a simplified approach to creating GUIs for
command-line programs. It's designed so _anyone_ can easily write their
own GUI, without having any programming experience.

It assumes your GUI is a direct frontend to one or more command-line programs,
with each command-line option having an associated GUI control widget. Several
kinds of widget are provided, for setting Filename, Color, Number or Font, or
for picking a Choice or Flag.

You probably know of a handful of command-line applications that would be
much better with a GUI, even a cheesy-looking Tkinter one. This module makes
it easy to create one.


DEFINE CONTROLS

Say, if you have a program that takes input and output filenames:

    $ tovid -in movie.avi -out movie_encoded

then you can create GUI widgets for those options like this:

    Filename('-in', "Input filename")
    Filename('-out', "Output prefix")

These have the general format:

    Control('-option', "Label", ...)

where:

    Control   is a Control subclass, such as Filename, Choice, or Number,
              describing what type of value is being controlled;
    'option'  is a command-line option (without the leading '-'),
              whose value is set by the Control; and
    "Label"   is the text that should appear next to the GUI Control.

Other parameters include default value, help/tooltip text to show, allowable
values, and hints about how to draw the GUI control widget; they are specific
to the flavor of Control being used. Here's a sampling:

    Choice    Multiple-choice values
    Color     Color selection button
    Filename  Type or [Browse] a filename
    FileList  Add/remove/rearrange filename list
    Flag      Check box, yes or no
    Font      Font selection button
    List      Space-separated list
    Number    Number between A and B
    Text      Plain old text

For a full list of available Control subclasses and how to use them,
see libtovid/metagui/control.py.


CREATING THE GUI

First, give your Controls a place to live, in a Panel:

    general = Panel("General",
        Filename('-bgaudio', "Background audio file"),
        Flag('-submenus', "Create submenus"),
        Number('-menu-length', "Length of menu (seconds)", 0, 120)
        )

This will create three GUI widgets in a Panel labeled "General": one for typing
or browsing to a filename, one for enabling or disabling submenus, and another 
for setting menu length to a number between 0 and 120. You can nest panels
inside one another for grouping; sub-Panels have their own label and list of
Controls or sub-Panels.

Once you have a Panel, you can create an Application:

    app = Application('todisc', [general])

This says your application will run the 'todisc' command-line program,
passing options set by the "General" panel. Now, create the GUI:

    gui = GUI('MyGUI', [app])
    gui.run()

This creates the GUI, draws all the widgets, and will run your command-line
program at the push of a button.


CREATE A MULTI-PANEL GUI

If your program has a lot of options, one panel may not be enough to hold them
all without looking cluttered, so you may break them down into multiple Panels,
which will be shown in the GUI as tabs that you can switch between. Create
Panels like this:

    thumbs = Panel("Thumbnails",
        Color('-thumb-mist-color', ...),
        Text('-wave', ...)
    )
    text = Panel("Text",
        Font('-menu-font', ...),
        Number('-menu-fontsize', ...)
    )

then, create the Application and GUI:

    todisc = Application('todisc', [thumbs, text])
    gui = GUI('MyGUI', [todisc])
    gui.run()

If multiple panels are given to Application, a tabbed interface is created,
with one tab for each panel.


CREATE A MULTI-APPLICATION GUI

If your GUI needs to be able to run several different command-line programs,
you can create a multi-application GUI. Create panels for each application,
then create the applications:

    todisc = Application('todisc', [todisc_panel1, todisc_panel2])
    tovid = Application('tovid', [tovid_panel1, tovid_panel2])

and then the GUI:

    gui = GUI('MultiGUI', [todisc, tovid])
    gui.run()

See the apps/ directory for example GUIs.
"""

# Export everything from support, control, and gui modules
# (Open to suggestion about more elegant ways to do this...)
from support import *
from control import *
from gui import *
from support import __all__ as all_support
from control import __all__ as all_control
from gui import __all__ as all_gui
__all__ = [\
    'support',
    'control',
    'gui',
    'manpage',
    'builder',
    'tooltip',
    'widget'] + all_support + all_control + all_gui





#! /usr/bin/env python
# __init__.py

"""A wxPython GUI interface to tovid.

This GUI module has the following sub-modules:
    configs -- 
    constants --
    controls --
    dialogs --
    frames --
    icons --
    options --
    panels --
    util --
"""

__all__ = [\
    "configs", 
    "constants",
    "controls", 
    "dialogs", 
    "frames", 
    "icons", 
    "meta",
    "options", 
    "panels",
    "tooltip",
    "util"]



#! /usr/bin/env python
# encode.py

"""This module provides several backend for encoding video and audio.

One high-level function is provided:

    encode(infile, outfile, format, tvsys, method, ...)

where "..." is optional keyword arguments (described below). For example:

    encode('/video/foo.avi', '/video/bar.mpg', 'dvd', 'ntsc', 'mencoder')

This will encode '/video/foo.avi' to NTSC DVD format using mencoder, saving
the result as '/video/bar.mpg'. Keyword arguments may be used to further
refine encoding behavior, for example:

    encode('foo.avi', 'foo.mpg', 'dvd', 'pal', 'ffmpeg',
           quality=7, interlace='bottom', ...)

The supported keywords may vary by backend, but one of our goals should be to
implement a set of commonly-used keywords in all the backends.

Each backend has a top-level function:

    ffmpeg_encode(source, target, ...)
    mencoder_encode(source, target, ...)
    mpeg2enc_encode(source, target, ...)

The source and target are MediaFile objects from libtovid.media, containing
profiles of the input and output videos. Again, the "..." are optional keyword
arguments.
"""

__all__ = [\
    'encode',
    'get_encoder',
    'ffmpeg_encode',
    'encode_audio',
    'mencoder_encode',
    'mpeg2enc_encode']

import os
import math
import copy
import glob
from libtovid.cli import Command, Pipe
from libtovid.utils import float_to_ratio
from libtovid.transcode import rip
from libtovid.media import *
from libtovid.standard import fps
from libtovid import log

_bitrate_limits = {\
    'vcd': (1150, 1150),
    'kvcd': (400, 4000),
    'dvd-vcd': (400, 4000),
    'svcd': (900, 2400),
    'half-dvd': (600, 6000),
    'dvd': (900, 9000)
    }

# --------------------------------------------------------------------------
#
# Primary interface
#
# --------------------------------------------------------------------------

def encode(infile, outfile, format='dvd', tvsys='ntsc', method='ffmpeg',
           **kw):
    """Encode a multimedia file according to a target profile, saving the
    encoded file to outfile.
    
        infile:  Input filename
        outfile: Desired output filename (.mpg implied)
        format:  One of 'vcd', 'svcd', 'dvd' (case-insensitive)
        tvsys:   One of 'ntsc', 'pal' (case-insensitive)
        method:  Encoding backend: 'ffmpeg', 'mencoder', or 'mpeg2enc'
        **kw:    Additional keyword arguments (name=value)
    
    The supported keyword arguments vary by encoding method. See the encoding
    functions for what is available in each.
    """
    source = load_media(infile)
    # Add .mpg to outfile if not already present
    if not outfile.endswith('.mpg'):
        outfile += '.mpg'
    # Get an appropriate encoding target
    target = standard_media(format, tvsys)
    target.filename = outfile
    # Set desired aspect ratio, or auto
    if 'aspect' in kw:
        target = correct_aspect(source, target, kw['aspect'])
    else:
        target = correct_aspect(source, target, 'auto')
    
    # Some friendly output
    print "Source media:"
    print source
    print
    print "Target media:"
    print target
    
    # Get the appropriate encoding backend
    encode_method = get_encoder(method)
    # Evaluate high-level keywords
    kw = eval_keywords(**kw)
    # Encode!
    encode_method(source, target, **kw)


def get_encoder(backend):
    """Get an encoding function."""
    if backend == 'ffmpeg':
        return ffmpeg_encode
    elif backend == 'mencoder':
        return mencoder_encode
    elif backend == 'mpeg2enc':
        return mpeg2enc_encode

def eval_keywords(**kw):
    """Interpret keywords that affect other keywords, and return the result.
    These are keywords that can be shared between multiple encoding backends.

    Supported keywords:
    
        quality:    From 1 (lowest) to 10 (highest) video quality.
                    Overrides 'quant' and 'vbitrate' keywords.
        fit:        Size in MiB to fit output video to (overrides 'quality')
                    Overrides 'quant' and 'vbitrate' keywords.

    """
    if 'quality' in kw:
        kw['quant'] = 13-kw['quality']
        max_bitrate = _bitrate_limits[target.format][1]
        kw['vbitrate'] = kw['quality'] * max_bitrate / 10
    if 'fit' in kw:
        kw['quant'], kw['vbitrate'] = _fit(source, target, kw['fit'])
    return kw


# --------------------------------------------------------------------------
#
# ffmpeg backend
#
# --------------------------------------------------------------------------

def ffmpeg_encode(source, target, **kw):
    """Encode a multimedia video using ffmpeg.

        source:  Input MediaFile
        target:  Output MediaFile
        kw:      name=value keyword arguments
    
    Supported keywords:
    
        quant:      Minimum quantization, from 1-31 (1 being fewest artifacts)
        vbitrate:   Maximum video bitrate, in kilobits per second.
        abitrate:   Audio bitrate, in kilobits per second
        interlace:  'top' or 'bottom', to do interlaced encoding with
                    top or bottom field first

    For example:
    
        ffmpeg_encode(source, target, quant=4, vbitrate=7000)
    """
    # Build the ffmpeg command
    cmd = Command('ffmpeg')
    cmd.add('-i', source.filename)
    if target.format in ['vcd', 'svcd', 'dvd']:
        cmd.add('-tvstd', target.tvsys,
                '-target', '%s-%s' % (target.tvsys, target.format))

    # Interpret keyword arguments
    if 'quant' in kw:
        cmd.add('-qmin', kw['quant'], '-qmax', 31)
    if 'vbitrate' in kw:
        cmd.add('-b', '%dk' % kw['vbitrate'])
    if 'abitrate' in kw:
        cmd.add('-ab', '%dk' % kw['abitrate'])
    if 'interlace' in kw:
        if kw['interlace'] == 'bottom':
            cmd.add('-top', 0, '-flags', '+alt+ildct+ilme')
        elif kw['interlace'] == 'top':
            cmd.add('-top', 1, '-flags', '+alt+ildct+ilme')

    # Convert frame rate and audio sampling rate
    cmd.add('-r', target.fps,
            '-ar', target.samprate)
    
    # Convert scale/expand to ffmpeg's padding system
    if target.scale:
        cmd.add('-s', '%sx%s' % target.scale)
    if target.expand != target.scale:
        e_width, e_height = target.expand
        s_width, s_height = target.scale
        h_pad = (e_width - s_width) / 2
        v_pad = (e_height - s_height) / 2
        if h_pad > 0:
            cmd.add('-padleft', h_pad, '-padright', h_pad)
        if v_pad > 0:
            cmd.add('-padtop', v_pad, '-padbottom', v_pad)
    if target.widescreen:
        cmd.add('-aspect', '16:9')
    else:
        cmd.add('-aspect', '4:3')
    # Overwrite existing output files
    cmd.add('-y')
    cmd.add(target.filename)
    
    cmd.run()


# --------------------------------------------------------------------------
#
# mencoder backend
#
# --------------------------------------------------------------------------

def mencoder_encode(source, target, **kw):
    """Encode a multimedia video using mencoder.

        source:  Input MediaFile
        target:  Output MediaFile
        kw:      name=value keyword arguments
    
    Supported keywords:
    """

    # Build the mencoder command
    cmd = Command('mencoder')
    cmd.add(source.filename,
            '-o', target.filename,
            '-oac', 'lavc',
            '-ovc', 'lavc',
            '-of', 'mpeg')
    # Format
    cmd.add('-mpegopts')
    if target.format in ['vcd', 'svcd']:
        cmd.add('format=x%s' % target.format)
    else:
        cmd.add('format=dvd')
    
    # TODO: this assumes we only have ONE audio track.
    if source.has_audio:
        # Audio settings
        # Adjust sampling rate
        # TODO: Don't resample unless needed
        if source.samprate != target.samprate:
            log.info("Resampling needed to achieve %d Hz" % target.samprate)
            cmd.add('-srate', target.samprate)
            cmd.add('-af', 'lavcresample=%s' % target.samprate)
        else:
            log.info("No resampling needed, already at %d Hz" % target.samprate)
        
    else:
        log.info("No audio file, generating silence of %f seconds." % \
                 source.length)
        # Generate silence.
        if target.format in ['vcd', 'svcd']:
            audiofile = '%s.mpa' % target.filename
        else:
            audiofile = '%s.ac3' % target.filename
        encode_audio(source, audiofile, target)
        # TODO: make this work, it's still not adding the ac3 file correctly.
        cmd.add('-audiofile', audiofile)

    # Video codec
    if target.format == 'vcd':
        lavcopts = 'vcodec=mpeg1video'
    else:
        lavcopts = 'vcodec=mpeg2video'
    # Audio codec
    lavcopts += ':acodec=%s' % target.acodec
    lavcopts += ':abitrate=%s:vbitrate=%s' % \
            (target.abitrate, target.vbitrate)
    # Maximum video bitrate
    lavcopts += ':vrc_maxrate=%s' % target.vbitrate
    if target.format == 'vcd':
        lavcopts += ':vrc_buf_size=327'
    elif target.format == 'svcd':
        lavcopts += ':vrc_buf_size=917'
    else:
        lavcopts += ':vrc_buf_size=1835'
    # Set appropriate target aspect
    if target.widescreen:
        lavcopts += ':aspect=16/9'
    else:
        lavcopts += ':aspect=4/3'
    # Pass all lavcopts together
    cmd.add('-lavcopts', lavcopts)

    # FPS
    if target.tvsys == 'pal':
        cmd.add('-ofps', '25/1')
    elif target.tvsys == 'ntsc':
        cmd.add('-ofps', '30000/1001') # ~= 29.97

    # Scale/expand to fit target frame
    if target.scale:
        vfilter = 'scale=%s:%s' % target.scale
        # Expand is not done unless also scaling
        if target.expand != target.scale:
            vfilter += ',expand=%s:%s' % target.expand
        cmd.add('-vf', vfilter)

    cmd.run()

# --------------------------------------------------------------------------
#
# mpeg2enc backend
#
# --------------------------------------------------------------------------


def mpeg2enc_encode(source, target, **kw):
    """Encode a multimedia video using mplayer|yuvfps|mpeg2enc.
    
        source:  Input MediaFile
        target:  Output MediaFile

    """
    log.warning("This encoder is very experimental, and may not work.")

    outname = target.filename
    # YUV raw video FIFO, for piping video from mplayer to mpeg2enc
    yuvfile = '%s.yuv' % outname
    try:
        os.remove(yuvfile)
    except:
        pass
    os.mkfifo(yuvfile)
    
    # Filenames for intermediate streams (ac3/m2v etc.)
    # Appropriate suffix for audio stream
    if target.format in ['vcd', 'svcd']:
        audiofile = '%s.mp2' % outname
    else:
        audiofile = '%s.ac3' % outname
    # Appropriate suffix for video stream
    if target.format == 'vcd':
        videofile = '%s.m1v' % outname
    else:
        videofile = '%s.m2v' % outname
    # Do audio
    encode_audio(source, audiofile, target)
    # Do video
    rip.rip_video(source, yuvfile, target)
    encode_video(source, yuvfile, videofile, target)
    # Combine audio and video
    mplex_streams(videofile, audiofile, target)


def encode_audio(source, audiofile, target):
    """Encode the audio stream in a source file to a target format, saving
    to the given filename.

        source:    Input MediaFile
        audiofile: Filename for encoded audio
        target:    Output MediaFile

    If no audio is present in the source file, encode silence.
    """
    cmd = Command('ffmpeg')

    # If source has audio, encode it
    if source.has_audio:
        cmd.add('-i', source.filename)
    # Otherwise, generate 4-second silence
    else:
        # Add silence the length of source length
        ln = source.length
        #print "Source.length: %f" % ln
        if ln < 4:
            # Minimum 4 secs :)
            ln = 4.0
        cmd.add('-f', 's16le', '-i', '/dev/zero', '-t', '%f' % ln)
    cmd.add('-vn', '-ac', '2', '-ab', '224')
    cmd.add('-ar', target.samprate)
    cmd.add('-acodec', target.acodec)
    cmd.add('-y', audiofile)

    cmd.run()


def encode_video(source, yuvfile, videofile, target):
    """Encode a yuv4mpeg stream to an MPEG video stream.
    
        source:    Input MediaFile
        yuvfile:   Filename of .yuv stream coming from mplayer
        videofile: Filename of .m[1|2]v to write encoded video stream to
        target:    Output MediaFile
        
    """
    # TODO: Control over quality (bitrate/quantization) and disc split size,
    # corresp. to $VID_BITRATE, $MPEG2_QUALITY, $DISC_SIZE, etc.
    # Missing options (compared to tovid)
    # -S 700 -B 247 -b 2080 -v 0 -4 2 -2 1 -q 5 -H -o FILE
    pipe = Pipe()

    # Feed the .yuv file into the pipeline
    cat = Command('cat', yuvfile)
    pipe.add(cat)

    # Adjust framerate if necessary by piping through yuvfps
    if source.fps != target.fps:
        log.info("Adjusting framerate")
        yuvfps = Command('yuvfps')
        yuvfps.add('-r', float_to_ratio(target.fps))
        pipe.add(yuvfps)

    # Encode the resulting .yuv stream by piping into mpeg2enc
    mpeg2enc = Command('mpeg2enc')
    # TV system
    if target.tvsys == 'pal':
        mpeg2enc.add('-F', '3', '-n', 'p')
    elif target.tvsys == 'ntsc':
        mpeg2enc.add('-F', '4', '-n', 'n')
    # Format
    format = target.format
    if format == 'vcd':
        mpeg2enc.add('-f', '1')
    elif format == 'svcd':
        mpeg2enc.add('-f', '4')
    elif 'dvd' in format:
        mpeg2enc.add('-f', '8')
    # Aspect ratio
    if target.widescreen:
        mpeg2enc.add('-a', '3')
    else:
        mpeg2enc.add('-a', '2')
    mpeg2enc.add('-o', videofile)
    pipe.add(mpeg2enc)
    
    # Run the pipeline to encode the video stream
    pipe.run()


def mplex_streams(vstream, astream, target):
    """Multiplex audio and video stream files to the given target.
    
        vstream:  Filename of MPEG video stream
        astream:  Filename of MP2/AC3 audio stream
        target:   Profile of output file
        
    """
    cmd = Command('mplex')
    format = target.format
    if format == 'vcd':
        cmd.add('-f', '1')
    elif format == 'dvd-vcd':
        cmd.add('-V', '-f', '8')
    elif format == 'svcd':
        cmd.add('-V', '-f', '4', '-b', '230')
    elif format == 'half-dvd':
        cmd.add('-V', '-f', '8', '-b', '300')
    elif format == 'dvd':
        cmd.add('-V', '-f', '8', '-b', '400')
    # elif format == 'kvcd':
    #   cmd += ' -V -f 5 -b 350 -r 10800 '
    cmd.add(vstream, astream, '-o', target.filename)
    # Run the command to multiplex the streams
    cmd.run()

# --------------------------------------------------------------------------
#
# Frame encoder
#
# --------------------------------------------------------------------------

def encode_frames(imagedir, outfile, format, tvsys, aspect, interlaced=False):
    """Convert an image sequence in the given directory to match a target
    MediaFile, putting the output stream in outfile.

        imagedir:   Directory containing images (and only images)
        outfile:    Filename for output.
        tvsys:      TVsys desired for output (to deal with FPS)
        aspect:     Aspect ratio ('4:3', '16:9')
        interlaced: Frames are interlaced material.
        
    Currently supports JPG and PNG images; input images must already be
    at the desired target resolution.
    """
    #
    # Trying to implement interlaced in encoding failed with:
    #    ++ WARN: [mpeg2enc] Frame height won't split into two equal field pictures...
    #    ++ WARN: [mpeg2enc] forcing encoding as progressive video
    # in all cases. Setting -Ip -L1 with pnv2yuv and -I2 in mpeg2enc, etc..
    # The option is still there, so that compatibility continues, but it
    # currently does nothing.
    #
    
    # Use absolute path name
    imagedir = os.path.abspath(imagedir)
    print "Creating video stream from image sequence in %s" % imagedir
    # Determine image type
    images = glob.glob("%s/*" % imagedir)
    extension = images[0][-3:]
    if extension not in ['jpg', 'png']:
        raise ValueError, "Image format '%s' isn't currently supported to "\
              "render video from still frames" % extension
    # Make sure remaining image files have the same extension
    for img in images:
        if not img.endswith(extension):
            raise RuntimeWarning, "%s does not have a .%s extension" %\
                  (img, extension)

    pipe = Pipe()

    # Use jpeg2yuv/png2yuv to stream images
    if extension == 'jpg':
        jpeg2yuv = Command('jpeg2yuv')
        
        jpeg2yuv.add('-Ip') # Progressive
            
        jpeg2yuv.add('-f', '%.3f' % fps(tvsys),
                     '-j', '%s/%%08d.%s' % (imagedir, extension))
        pipe.add(jpeg2yuv)
    elif extension == 'png':
        #ls = Command('sh', '-c', 'ls %s/*.png' % imagedir)
        #xargs = Command('xargs', '-n1', 'pngtopnm')
        png2yuv = Command('png2yuv')

        png2yuv.add('-Ip') # Progressive
            

        png2yuv.add('-f', '%.3f' % fps(tvsys),
                    '-j', '%s/%%08d.png' % (imagedir))
        
        pipe.add(png2yuv)

        #pipe.add(ls, xargs, png2yuv)
        #cmd += 'pnmtoy4m -Ip -F %s %s/*.png' % standard.fpsratio(tvsys)

    # TODO: Scale to correct target size using yuvscaler or similar
    
    # Pipe image stream into mpeg2enc to encode
    mpeg2enc = Command('mpeg2enc')

    # Anyways.
    mpeg2enc.add('-I 0') # Progressive

    # Bottom-first, determined in render/drawing.py::interlace_drawings()
    if (interlaced):
        mpeg2enc.add('--playback-field-order', 'b')  # Bottom-first field order
        # conforming to drawing.py::interlace_drawings()
    
    mpeg2enc.add('-v', '0',
                 '-q' '3',
                 '-o' '%s' % outfile)
    # Format
    if format == 'vcd':
        mpeg2enc.add('--format', '1')
    elif format == 'svcd':
        mpeg2enc.add('--format', '4')
    else:
        mpeg2enc.add('--format', '8')
    # Aspect ratio
    if aspect == '16:9':
        mpeg2enc.add('-a', '3')
    elif aspect == '4:3':
        mpeg2enc.add('-a', '2')

    pipe.add(mpeg2enc)

    pipe.run(capture=True)
    pipe.get_output() # and wait :)


def _fit(source, target, fit_size):
    """Return video (quantization, bitrate) to fit a video into a given size.
    
        source:   MediaFile input (the video being encoded)
        target:   MediaFile output (desired target profile)
        fit_size: Desired encoded file size, in MiB

    """
    assert isinstance(source, MediaFile)
    assert isinstance(target, MediaFile)
    fit_size = float(fit_size)
    mpeg_overhead = fit_size / 100
    aud_vid_size = fit_size - mpeg_overhead
    audio_size = float(source.length * target.abitrate) / (8*1024)
    video_size = aud_vid_size - audio_size
    vid_bitrate = int(video_size*8*1024 / source.length)

    print "Length:", source.length, "seconds"
    print "Overhead:", mpeg_overhead, "MiB"
    print "Audio:", audio_size, "MiB"
    print "Video:", video_size, "MiB"
    print "VBitrate:", vid_bitrate, "kbps"

    # Keep bitrates sane for each disc format
    lower, upper = _bitrate_limits[target.format]
    quant = 3
    if vid_bitrate < lower:
        return (quant, lower)
    elif vid_bitrate > upper:
        return (quant, upper)
    else:
        return (quant, vid_bitrate)




#! /usr/bin/env python
# -=- encoding: latin-1 -=-
# drawing.py

# Run this script standalone for a demonstration:
#    $ python libtovid/render/drawing.py

# To use this module from your Python interpreter, run:
#    $ python
#    >>> from libtovid.render.drawing import *

"""Interactive vector drawing and rendering interface.

This module provides a class called Drawing, which provides a blank canvas
that can be painted with drawing functions. To use it, simply provide a
width and height, in arbitrary units:

    >>> drawing = Drawing(800, 600)   # width, height

This does not imply the *pixel* resolution of your drawing. Remember, this
is vector drawing, with (virtually) infinite precision; all you're describing
here is the *shape* of the drawing--800 units by 600 units, which implies a
4:3 aspect ratio. You could just as well do Drawing(4, 3) or Drawing(80, 60);
the only thing that changes is the coordinate system you want to use for
drawing operations.

Back on topic: you can now draw shapes on the drawing, fill and stroke them:

    >>> drawing.circle(500, 300, 150)        # (x, y, radius)
    >>> drawing.fill('blue', 0.8)            # color name with optional alpha
    >>> drawing.stroke('rgb(0, 128, 255)')   # rgb values 0-255

Then add more shapes:

    >>> drawing.rectangle(25, 25, 320, 240)  # x, y, width, height
    >>> drawing.fill('rgb(40%, 80%, 10%)')   # rgb percentages
    >>> drawing.stroke('#FF0080', 0.5)       # rgb hex notation with an alpha

To see what the drawing looks like so far, do:

    >>> display(drawing)                     # display at existing size
    >>> display(drawing, 720, 480)           # display at different size
    >>> display(drawing, fix_aspect=True)    # display in correct aspect ratio

You may then decide to add more to the drawing:

    >>> drawing.set_source('white')
    >>> drawing.text("Dennis", 50, 100)      # text, x, y

And preview again:

    >>> display(drawing)

Once you finish your beautiful painting, save it as a nice high-res PNG!

    >>> save_png(drawing, "my.png", 1600, 1200)
    
Cairo references:

    [1] http://www.cairographics.org
    [2] http://tortall.net/mu/wiki/CairoTutorial

"""

__all__ = [
    'Drawing',
    'render',
    'display',
    'save_image',
    'save_jpg',
    'save_pdf',
    'save_png',
    'save_ps',
    'save_svg',
    'write_ppm']

import os
import sys
import time
import commands
from copy import copy
from math import pi, sqrt
import cairo
import Image      # for JPG export
import ImageColor # for getrgb, getcolor
import ImageFile  # for write_png()
from libtovid import log
from libtovid.utils import to_unicode

### --------------------------------------------------------------------
### Internal functions
### --------------------------------------------------------------------

def get_surface(width, height, surface_type='image', filename=''):
    """Get a cairo surface at the given dimensions.
    """
    if type(width) != int or type(height) != int:
        log.warning("Surface width and height should be integers."\
                    " Converting to int.")

    if surface_type == 'image':
        return cairo.ImageSurface(cairo.FORMAT_ARGB32,
                                  int(width), int(height))
    elif surface_type == 'svg':
        return cairo.SVGSurface(filename, width, height)
    elif surface_type == 'pdf':
        return cairo.PDFSurface(filename, width, height)
    elif surface_type == 'ps':
        return cairo.PSSurface(filename, width, height)



### --------------------------------------------------------------------
### Classes
### --------------------------------------------------------------------

class Step:
    """An atomic drawing procedure, consisting of a closed function that
    draws on a given cairo surface, and human-readable information about
    what parameters were given to it.
    """
    def __init__(self, function, *args):
        self.function = function
        self.name = function.__name__.lstrip('_')
        self.args = args
    def __str__(self):
        return "%s%s" % (self.name, self.args)

# Drawing class notes:
#
# The Drawing class has a number of methods (circle, rectangle, fill, stroke
# and many others) that need to operate on a Cairo surface. But we'd like to
# delay execution of actually drawing on that surface--otherwise, we can't
# easily render a given Drawing to a custom resolution.
#
# Closures save the day here--that is, functions without "free variables".
# Anytime you "paint" on the Drawing, what's actually happening is a new
# function is getting created, whose sole purpose in life is to carry out
# that specific paint operation. These tiny, single-purpose functions are
# then added to a list of steps (self.steps) that will actually be
# executed at rendering-time (i.e., when you do display() or save_png).
#
# This not only lets us render a Drawing to different resolutions, but allows
# the possibility of rendering to different Cairo surfaces.

class Drawing:

    tk = None # tkinter widget to send redraws to
    
    def __init__(self, width=800, height=600, autodraw=False):
        """Create a blank drawing at the given size.
        
            width, height: Dimensions of the drawable region,
                           in arbitrary units
            autodraw:      Redraw a preview image after each
                           drawing operation
        """
        self.size = (width, height)
        self.w = width
        self.h = height
        self.aspect = float(width) / height
        self.autodraw = autodraw
        #self.commands = []
        self.steps = []
        # "Dummy" surface/context at the original resolution;
        # should not be drawn on
        self.surface = get_surface(width, height, 'image')
        self.cr = cairo.Context(self.surface)


    def addStep(self, func, *args):
        """Add a Step to self.steps using the given function and args."""
        step = Step(func, *args)
        self.steps.append(step)
        if self.autodraw and step.name in ['fill', 'stroke']:
            display(self, 640, 480, True)

    def doStep(self, func, *args):
        """Add the given Step, and execute it."""
        self.addStep(func, *args)
        func(self.cr)

    def history(self):
        """Return a formatted string of all steps in this Drawing."""
        result = ''
        for number, step in enumerate(self.steps):
            result += '%d. %s\n' % (number, step)
        return result


    ### -----------------------------------------------------------------
    ### Drawing commands
    ### -----------------------------------------------------------------

    def affine(self, rot_x, rot_y, scale_x, scale_y, translate_x, translate_y):
        """Define a 3x3 transformation matrix:
            
            [ scale_x  rot_y    translate_x ]
            [ rot_x    scale_y  translate_y ]
            [ 0        0        1           ]

        This is scaling, rotation, and translation in a single matrix,
        so it's a compact way to represent any transformation.
        
        See [http://www.w3.org/TR/SVG11/coords.html] for more on
        these matrices and how to use them.
        """
        mtx = cairo.Matrix(scale_x, rot_x, rot_y, scale_y,
                           translate_x, translate_y)
        def _affine(cr):
            cr.set_matrix(mtx)
        self.doStep(_affine, rot_x, rot_y, scale_x, scale_y,
                     translate_x, translate_y)


    def arc(self, x, y, radius, start_deg, end_deg):
        """Draw an arc from (x, y), with the specified radius, starting at
        degree start_deg and ending at end_deg.
        """
        def _arc(cr):
            cr.arc(x, y, radius, start_deg * pi/180., end_deg * pi/180.)
        self.doStep(_arc, x, y, radius, start_deg, end_deg)


    def arc_rad(self, x, y, radius, start_rad, end_rad):
        """Draw an arc from (x, y), with the specified radius, starting at
        radian start_rad and ending at end_rad.
        """
        def _arc_rad(cr):
            cr.arc(x, y, radius, start_rad, end_rad)
        self.doStep(_arc_rad, x, y, radius, start_rad, end_rad)


    # TODO: Rewrite/cleanup bezier function
    def bezier(self, points, rel=False, close_path=False):
        """Create a Bézier curve.

            points should look like:
            
            points = [[(x0, y0), (x_ctl1, y_ctl1), (x_ctl2, y_ctl2)],
                      [(x0, y0), (x_ctl1, y_ctl1), (x_ctl2, y_ctl2)],
                      [(x0, y0), (x_ctl1, y_ctl1), (x_ctl2, y_ctl2)],
                      [(x0, y0)],
                      ]
             
        where (x0, y0) are the point coordinates, and x_ctl* are the control
        points coordinates.

            *ctl1 - control from last point
            *ctl2 - control to next point

        The last line would specify that control points
        are at the same place (x0, y0), drawing straight lines.

            rel -- when this is True, the x_ctl and y_ctl become relatives
                   to the (x0, y0) points
            close_path -- set this to True to call cr.close_path() before
                          stroking.

        """
        assert len(points) > 1, "You need at least two points"
        for pt in points:
            # If only one point is specified, copy it for control points
            if len(pt) == 1:
                pt.append(pt[0])       # add the two identical
                pt.append(pt[0])       # control points
            assert len(pt) == 3, "You need to specify three tuples for each point, or one single"
            for y in pt:
                assert isinstance(y, tuple) and len(y) == 2, "You need "\
                       "to specify two-values tuples"
        # Map relative stuff to absolute, when rel=True
        if rel:
            for x in range(0,len(points)):
                pt = points[x]
                assert len(pt) == 3, "In relative mode, you must "\
                       "specify control points"
                # Render relative values to absolute values.
                npt = []
                #             #  x      #  y
                npt.append([pt[0][0], pt[0][1]]) # x0, y0
                npt.append([pt[0][0] + pt[1][0], pt[0][1] + pt[1][1]])
                npt.append([pt[0][0] + pt[2][0], pt[0][1] + pt[2][1]])
                points[x] = npt
                    
        # Define the actual drawing function
        def _bezier(cr):
            # Move to the first x,y in the first point
            cr.new_path()
            cr.move_to(points[0][0][0], points[0][0][1])
            for pt in points:
                cr.curve_to(pt[2][0], pt[2][1],
                            pt[0][0], pt[0][1],
                            pt[1][0], pt[1][1])
            if close_path:
                cr.close_path()
        self.doStep(_bezier, points, rel, close_path)


    def circle(self, center_x, center_y, radius):
        """Draw a circle defined by center point and radius."""
        def _circle(cr):
            cr.new_path()
            cr.arc(center_x, center_y, radius, 0, 2*pi)
        self.doStep(_circle, center_x, center_y, radius)

    # TODO: add clip stuff...


    def fill(self, source=None, opacity=None):
        """Fill the current (closed) path with an optionally given color.

        If arguments are present, they are passed to set_source()
        before filling. Note that this source will be applied to further
        fill()ings or stroke()ings or text() calls.
        """
        # Optionally set fill color, and save it.
        if source is not None:
            self.set_source(source, opacity)

        def _fill(cr):
            cr.fill_preserve()
        self.doStep(_fill, source, opacity)


    def fill_rule(self, rule):
        """Set the fill rule to one of:
        
            evenodd, winding (nonzero)

        This determines which parts of an overlapping path polygon will
        be filled with the fill() command.
            
        http://www.w3.org/TR/SVG/painting.html#FillRuleProperty
        """
        tr = {'evenodd': cairo.FILL_RULE_EVEN_ODD,
              'winding': cairo.FILL_RULE_WINDING}

        # Test value
        tr[rule]
        
        def _fill_rule(cr):
            cr.set_fill_rule(tr[rule])
        self.doStep(_fill_rule, rule)


    def font(self, name, slant='normal', weight='normal'):
        """Set the current font, by name.

        name -- name of the Font, or family (sans-serif, serif)
        slant -- one of: italic, normal, oblique
        weight -- one of: normal, bold
        """
        sl = {'italic': cairo.FONT_SLANT_ITALIC,
              'normal': cairo.FONT_SLANT_NORMAL,
              'oblique': cairo.FONT_SLANT_OBLIQUE}
        wg = {'normal': cairo.FONT_WEIGHT_NORMAL,
              'bold': cairo.FONT_WEIGHT_BOLD}
        def _font(cr):
            cr.select_font_face(name, sl[slant], wg[weight])
        self.doStep(_font, name, slant, weight)


    def font_size(self, pointsize):
        """Set the current font size in points."""
        def _font_size(cr):
            cr.set_font_size(pointsize)
        self.doStep(_font_size, pointsize)


    def font_stretch(self, x=1.0, y=1.0):
        """Set the font stretch type in both directions to one of:

        = 1.0 -- normal
        > 1.0 -- strench
        < 1.0 -- shrink

        """
        def _font_stretch(cr):
            m = cr.get_font_matrix()
            m.scale(x, y)
            cr.set_font_matrix(m)
        self.doStep(_font_stretch, x, y)


    def font_rotate(self, degrees):
        """Set the font rotation, in degrees.
        """
        def _font_rotate(cr):
            m = cr.get_font_matrix()
            m.rotate(degrees * pi/180.)
            cr.set_font_matrix(m)
        self.doStep(_font_rotate, degrees)

    def image_surface(self, x, y, width, height, surface, mask=None):
        """Draw a given cairo.ImageSurface centered at (x, y), at the given
        width and height.

        If you specify mask, it must be a cairo ImageSurface.
        
        """
        # Calculate centering and scaling
        add_x, add_y = (0, 0)
        dw = float(width) / float(surface.get_width())
        dh = float(height) / float(surface.get_height())
        if (dw > dh):
            scale = dh
            add_x = (width - dh * float(surface.get_width())) / 2
        else:
            scale = dw
            add_y = (height - dw * float(surface.get_height())) / 2
        
        # Save context and get the surface to right dimensions
        self.save()
        self.translate(x + add_x, y + add_y)
        self.scale(scale, scale)

        # Create and append the drawing function
        if (mask):
            def _image_surface(cr):
                cr.set_source_surface(surface)
                cr.mask_surface(mask)
            self.doStep(_image_surface, x, y, width, height, surface, mask)
        else:
            def _image_surface(cr):
                cr.set_source_surface(surface)
                cr.paint()
            self.doStep(_image_surface, x, y, width, height, surface, mask)

        self.restore()
    

    def image(self, x, y, width, height, source):
        """Draw an image centered at (x, y), scaled to the given width and
        height. Return the corresponding cairo.ImageSurface object.

        source -- a .png filename (quicker and alpha present),
                  a cairo.ImageSurface object, which is even better,
                  a file object
                  a filename - any file supported by python-imaging,
                               a.k.a PIL [1]

        Ref:
          [1] http://www.pythonware.com/library/pil/handbook/formats.htm

        You can apply some operator() to manipulation how the image is going
        to show up. See operator()
        """
        # Determine what type of source was given, and make it into
        # a cairo.ImageSurface
        if isinstance(source, cairo.ImageSurface):
            surface = source
        # PNG files can be added directly
        elif isinstance(source, str) and source.endswith('.png'):
            surface = cairo.ImageSurface.create_from_png(source)
        # Other formats must be converted to PNG
        else:
            infile = Image.open(source)
            outfile = open('/tmp/export.png', 'wb+')
            infile.save(outfile, 'PNG')
            outfile.seek(0)
            surface = cairo.ImageSurface.create_from_png(outfile)
            outfile.close()
        # Let someone else do the dirty work
        self.image_surface(x, y, width, height, surface)

        return surface


    def line(self, x0, y0, x1, y1):
        """Set new path as a line from (x0, y0) to (x1, y1).

        Don't forget to stroke()
        """
        def _line(cr):
            cr.new_path()
            cr.move_to(x0, y0)
            cr.line_to(x1, y1)
        self.doStep(_line, x0, y0, x1, y1)


    def operator(self, operator='clear'):
        """Set the operator mode.

        operator -- One of: clear
                            source, over, in, out, atop
                            dest, dest_over, dest_in, dest_out, dest_atop
                            xor, add, saturate

        """
        ops = {'clear': cairo.OPERATOR_CLEAR,
               'source': cairo.OPERATOR_SOURCE,
               'over': cairo.OPERATOR_OVER,
               'in': cairo.OPERATOR_IN,
               'out': cairo.OPERATOR_OUT,
               'atop': cairo.OPERATOR_ATOP,
               'dest': cairo.OPERATOR_DEST,
               'dest_over': cairo.OPERATOR_DEST_OVER,
               'dest_in': cairo.OPERATOR_DEST_IN,
               'dest_out': cairo.OPERATOR_DEST_OUT,
               'dest_atop': cairo.OPERATOR_DEST_ATOP,
               'xor': cairo.OPERATOR_XOR,
               'add': cairo.OPERATOR_ADD,
               'saturate': cairo.OPERATOR_SATURATE,
               }

        # Test bad value
        ops[operator]

        def _operator(cr):
            cr.set_operator(ops[operator])
        self.doStep(_operator, operator)


    def paint(self):
        """Paint the current source everywhere within the current clip
        region."""
        def _paint(cr):
            cr.paint()
        self.doStep(_paint)


    def paint_with_alpha(self, alpha):
        """Paints the current source everywhere within the current clip
        region using a mask of constant alpha value alpha.

        The effect is similar to paint(), but the drawing is faded out
        using the alpha value."""
        def _paint_with_alpha(cr):
            cr.paint_with_alpha(alpha)
        self.doStep(_paint_with_alpha, alpha)

    def point(self, x, y):
        """Draw a point (really a tiny circle) at (x, y)."""
        # Circle radius 1/1000 the drawing width
        radius = float(self.size[0]) / 1000
        def _point(cr):
            cr.new_path()
            cr.arc(x, y, radius, 0, 2*pi)
        self.doStep(_point, x, y)


    def polygon(self, points, close_path=True):
        """Define a polygonal path defined by (x, y) points in the given
        list.

            points = [(x0, y0),
                       (x1, y1),
                       (x2, y2)]

        Draw strokes and filling yourself, with fill() and stroke().
        """
        nlist = []
        for tup in points:
            nlist.append([tup])
        self.bezier(nlist, False, close_path)

    
    def polyline(self, points, close_path=True):
        """Draw a polygon defined by (x, y) points in the given list.

        This is a short- (or long-) hand for polygon. In Cairo, you draw
        your filling (fill()), or strokes (stroke()) yourself, so
        having polyline and polygon is basiclly useless.
        """
        self.polygon(points, close_path)

    
    def rectangle_corners(self, x0, y0, x1, y1):
        """Draw a rectangle defined by opposite corners (x0, y0) and (x1, y1).
        """
        def _rectangle_corners(cr):
            cr.new_path()
            cr.rectangle(x0, y0, x1-x0, y1-y0)
        self.doStep(_rectangle_corners, x0, y0, x1, y1)


    def rectangle(self, x, y, width, height):
        """Draw a rectangle with top-left corner at (x, y), and (width x height)
        in size.
        """
        def _rectangle(cr):
            cr.new_path()
            cr.rectangle(x, y, width, height)
        self.doStep(_rectangle, x, y, width, height)


    def rotate_deg(self, degrees):
        """Rotate by the given number of degrees."""
        def _rotate_deg(cr):
            m = cr.get_matrix()
            m.rotate(degrees * pi/180.)
            cr.set_matrix(m)
        self.doStep(_rotate_deg, degrees)

    rotate = rotate_deg


    def rotate_rad(self, rad):
        """Rotate by the given number of radians."""
        def _rotate_rad(cr):
            m = cr.get_matrix()
            m.rotate(rad)
            cr.set_matrix(m)
        self.doStep(_rotate_rad, rad)


    def roundrectangle(self, x0, y0, x1, y1, bevel_width, bevel_height):
        """Draw a rounded rectangle from (x0, y0) to (x1, y1), with
        a bevel size of (bevel_width, bevel_height)."""
        bw = bevel_width
        bh = bevel_height
        # Add bezier points
        # Top left corner:
        tl1 = [(x0, y0 + bh), (0,0), (0,0)]
        tl2 = [(x0 + bw, y0), (0,0), (-bw,0)]
        tr1 = [(x1 - bw, y0), (0,0), (-bw,0)]
        tr2 = [(x1, y0 + bh), (0,0), (0,-bw)]
        br1 = [(x1, y1 - bh), (0,0), (0,-bh)]
        br2 = [(x1 - bw, y1), (0,0), (+bw,0)]
        bl1 = [(x0 + bw, y1), (0,0), (+bw,0)]
        bl2 = [(x0, y1 - bh), (0,0), (0,+bh)]
        end = [(x0, y0 + bh), (0,0), (0,0)]
        # Call in relative mode bezier control points.
        mylst = [tl1, tl2, tr1, tr2, br1, br2, bl1, bl2, end]
        # Let bezier do the work
        self.bezier(mylst, True)


    def set_source(self, source, opacity=None):
        """
        source -- One of the following color formats as a string or a tuple
                  (see below), another Surface or Surface-derivative object,
                  or a Pattern object (and its derivatives).
        opacity -- Opacity ranging from 0.0 to 1.0. Defaults to None. If
                   you specify a 'string' or 'tuple' RGB-like code as
                   source, opacity defaults to 1.0

        You can use the following *tuple* format to specify color:
        * (red, green, blue) -- with colors ranging from 0.0 to 1.0, like
                                Cairo wants them.

        You can use the following *string* formats to specify color:
        
        * Hexadecimal color specifiers, given as '#rgb' or '#rrggbb'.
          For example, '#ff0000' specifies pure red.
        * RGB functions, given as 'rgb(red, green, blue)' where the colour
          values are integers in the range 0 to 255. Alternatively, the color
          values can be given as three percentages (0% to 100%). For example,
          'rgb(255,0,0)' and 'rgb(100%,0%,0%)' both specify pure red.
        * Hue-Saturation-Lightness (HSL) functions, given as:
          'hsl(hue, saturation%, lightness%)' where hue is the colour given as
          an angle between 0 and 360 (red=0, green=120, blue=240), saturation
          is a value between 0% and 100% (gray=0%, full color=100%), and
          lightness is a value between 0% and 100% (black=0%, normal=50%,
          white=100%). For example, 'hsl(0,100%,50%)' is pure red.
        * Common HTML colour names. The ImageColor module provides some 140
          standard colour names, based on the colors supported by the X
          Window system and most web browsers. Colour names are case
          insensitive. For example, 'red' and 'Red' both specify pure red.
         
        """
        mysource = self.create_source(source, opacity)
        def _set_source(cr):
            cr.set_source(mysource)
        # Only do the step if the source could be created
        if mysource:
            self.doStep(_set_source, source, opacity)


    def create_source(self, source, opacity=None):
        """Return a source pattern (solid, gradient, image surface)
        with anything fed into it. Returns None if the source could not
        be created successfully.

        See the set_source() documentation for more details on inputs.

        You can feed anything returned by this function into set_source()
        """
        if isinstance(source, cairo.Pattern):
            return source
        # A string--color name, rgb(), hsv(), or hexadecimal
        elif isinstance(source, str):
            try:
                (r, g, b) = ImageColor.getrgb(source)
            except ValueError, err:
                log.error(err)
                return None
            alpha = 1.0
            if opacity is not None:
                alpha = opacity
            return cairo.SolidPattern(r / 255.0, g / 255.0,
                                      b / 255.0, alpha)
        # An (r, g, b) tuple
        elif isinstance(source, tuple):
            assert len(source) == 3
            (r, g, b) = source
            alpha = 1.0
            if opacity is not None:
                alpha = opacity
            return cairo.SolidPattern(r, g, b, alpha)
        # A cairo Gradient
        elif isinstance(source, cairo.Gradient):
            return source
        # A cairo Surface
        elif isinstance(source, cairo.Surface):
            return cairo.SurfacePattern(source)
        else:
            raise TypeError, "source must be one of: str, tuple, "\
                  "cairo.Pattern, cairo.Gradient or cairo.Surface"


    def get_source(self):
        """Return the actually used source pattern

        This returns either of these:

            cairo.SolidPattern (RGB or RGBA color)
            cairo.SurfacePattern (some other image)
            cairo.LinearGradient (linear gradient :)
            cairo.RadialGradient (radial gradient :)

        You can then use this source in set_source(), to use the same
        pattern for fill()ing or stroke()ing.
        """
        # TODO: Fix or remove this function
        return self.cr.get_source()


    def scale(self, factor_x, factor_y):
        """Set scaling to (factor_x, factor_y), where 1.0 means no scaling.

        Note that the center point for scaling is (0,0); to scale from a
        different center point, use scale_centered()
        """
        def _scale(cr):
            m = cr.get_matrix()
            m.scale(factor_x, factor_y)
            cr.set_matrix(m)
        self.doStep(_scale, factor_x, factor_y)


    def scale_centered(self, center_x, center_y, factor_x, factor_y):
        """Set scaling to (dx, dy), where dx and dy are horizontal and
        vertical ratios (floats).

        scale_centered() will keep the (center_x, center_y) in the middle
        of the scaling, operating a translation in addition to scaling.
        """
        def _scale_centered(cr):
            m = cr.get_matrix()
            # Ouch, we want to keep the center x,y at the same place?!
            m.translate(-(center_x * factor_x - center_x),
                        -(center_y * factor_y - center_y))
            m.scale(factor_x, factor_y)
            cr.set_matrix(m)
        self.doStep(_scale_centered, center_x, center_y, factor_x, factor_y)


    def stroke(self, source=None, opacity=None):
        """Stroke the current shape.

        See fill() documentation for parameter explanation. They are the
        same.
        """
        # Optionally set stroke color, and save it.
        if source is not None:
            self.set_source(source, opacity)

        def _stroke(cr):
            # Optionally set fill color, and save it.
            cr.stroke_preserve()
        self.doStep(_stroke, source, opacity)


    def stroke_antialias(self, do_antialias=True):
        """Enable/disable antialiasing for strokes.

            do_antialias: True, False, 'none', 'default', or 'gray'
        
        This does not affect text. See text_antialias() for more info
        on text antialiasing.
        """
        if do_antialias:
            aa_type = cairo.ANTIALIAS_GRAY
        else:
            aa_type = cairo.ANTIALIAS_NONE
            
        def _stroke_antialias(cr):
            cr.set_antialias(aa_type)
        self.doStep(_stroke_antialias, do_antialias)


    def stroke_dash(self, array, offset=0.0):
        """Set the dash style.

            array:  A list of floats, alternating between ON and OFF for
                    each value
            offset: An offset into the dash pattern at which the stroke
                    should start
        
        For example:
        
            >>> stroke_dash([5.0, 3.0])

        alternates between 5 pixels on, and 3 pixels off.
        """
        def _stroke_dash(cr):
            cr.set_dash(array, offset)
        self.doStep(_stroke_dash, array, offset)


    def stroke_linecap(self, cap_type):
        """Set the type of line cap to use for stroking.

            cap_type:  One of 'butt', 'round', or 'square'
        """
        dic = {'butt': cairo.LINE_CAP_BUTT,
               'round': cairo.LINE_CAP_ROUND,
               'square': cairo.LINE_CAP_SQUARE}

        # Test value
        dic[cap_type]
        
        def _stroke_linecap(cr):
            cr.set_line_cap(dic[cap_type])
        self.doStep(_stroke_linecap, cap_type)


    def stroke_linejoin(self, join_type):
        """Set the type of line-joining to do for stroking.

            join_type:  One of 'bevel', 'miter', or 'round'.
        """
        dic = {'bevel': cairo.LINE_JOIN_BEVEL,
               'miter': cairo.LINE_JOIN_MITER,
               'round': cairo.LINE_JOIN_ROUND}

        # Test value
        dic[join_type]
        
        def _stroke_linejoin(cr):
            cr.set_line_join(dic[join_type])
        self.doStep(_stroke_linejoin, join_type)


    def stroke_width(self, width):
        """Set the current stroke width in pixels."""
        # (Pixels or points?)
        def _stroke_width(cr):
            cr.set_line_width(width)
        self.doStep(_stroke_width, width)


    def text(self, text_string, x, y, align='left'):
        """Draw the given text string.

            text_string: utf8 encoded text string
            (x, y):      lower-left corner position
        
        Set the text's color with set_source() before calling text().
        """
        text_string = to_unicode(text_string)

        def _text(cr):
            (dx, dy, w, h, ax, ay) = cr.text_extents(text_string)
            if align == 'right':
                nx = x - w
            elif align == 'center':
                nx = x - w / 2
            else:
                nx = x
            cr.move_to(nx, y)
            cr.show_text(text_string)
        self.doStep(_text, text_string, x, y, align)


    def text_path(self, text_string, x, y):
        """Same as text(), but sets a path and doesn't draw anything.
        Call stroke() or fill() yourself.

        Has the parameters and returns the same values as text().

        Beware that text() calls the Cairo function show_text() which
        caches glyphs, so is more efficient when dealing with large
        texts.
        """
        if not isinstance(text_string, unicode):
            text_string = unicode(text_string.decode('latin-1'))
        def _text_path(cr):
            cr.move_to(x, y)
            cr.text_path(text_string)
        self.doStep(_text_path, text_string, x, y)


    def text_extents(self, text):
        """Returns the dimensions of the to-be-drawn text.

        Call this before calling text() if you want to place it well,
        according to the dimensions of the text rectangle to be drawn.

        See text()'s return value for details.

        Returns:
            (x_bearing, y_bearing, width, height, x_advance, y_advance)
        """
        text = to_unicode(text)
        return self.cr.text_extents(text)


    def translate(self, dx, dy):
        """Do a translation by (dx, dy) pixels."""
        def _translate(cr):
            m = cr.get_matrix()
            m.translate(dx, dy)
            cr.set_matrix(m)
        self.doStep(_translate, dx, dy)


    def push_group(self):
        """Temporarily redirects drawing to an intermediate surface
        known as a group.

        The redirection lasts until the group is completed by a call
        to pop_group() or pop_group_to_source(). These calls provide
        the result of any drawing to the group as a pattern, (either
        as an explicit object, or set as the source pattern).
        """
        def _push_group(cr):
            cr.push_group()
        self.doStep(_push_group)


    def pop_group_to_source(self):
        """Terminates the redirection begun by a call to push_group() and
        installs the resulting pattern as the source pattern in the current
        cairo context.
        """
        def _pop_group_to_source(cr):
            cr.pop_group_to_source()
        self.doStep(_pop_group_to_source)


    def save(self):
        """Save state for Cairo and local contexts.

        You can save in a nested manner, and calls to restore() will pop
        the latest saved context.
        """
        def _save(cr):
            cr.save()
        self.doStep(_save)
    BEGIN = save

    def restore(self):
        """Restore the previously saved context."""
        def _restore(cr):
            cr.restore()
        self.doStep(_restore)
    END = restore

    ### ----------------------------------------------------------------
    ### Untested methods
    ### ----------------------------------------------------------------
    
    
#    def font_slant(self, style):
#        """Set the font slant. Use one of:
#    
#           normal, italic, oblique
#        """
#        tr = {'italic': cairo.FONT_SLANT_ITALIC,
#              'normal': cairo.FONT_SLANT_NORMAL,
#              'oblique': cairo.FONT_SLANT_OBLIQUE}
#        def _font_slant(cr):
#            cr.set_slant(tr[style])
#        self.doStep(_font_slant, style)
#    # Alias
#    font_style = font_slant


#    def font_weight(self, weight):
#        """Set the font weight to one of:
#        
#            normal, bold
#        """
#        tr = {'normal': cairo.FONT_WEIGHT_NORMAL,
#              'bold': cairo.FONT_WEIGHT_BOLD}
#        def _font_weight(cr):
#            cr.set_font_weight(tr[weight])
#        self.doStep(_font_weight, weight)


    def font_family(self, family):
        """Set the current font, by family name."""
        def _font_family(cr):
            cr.select_font_face(family)
        self.doStep(_font_family, family)
    # Alias
    font_face = font_family


    def text_align(self, text_string, x, y, align='left'):
        """Draw the given text string.

            text_string: utf8 encoded text string
            (x, y):      lower-left corner position
            align:       'left', 'center', 'right'. This only changes the
                         alignment in 'x', and 'y' stays the baseline.

        Set the text's color with set_source() before calling text().
        """
        (dx, dy, w, h, ax, ay) = self.cr.text_extents(text_string)
        w = 0
        if align == 'right':
            x = x - w
        elif align == 'center':
            x = x - w / 2
        else: # 'left'
            x = x
        # Let text() do the drawing
        self.text(text_string, x, y)


    def text_path_align(self, x, y, text_string, centered=False):
        """Same as text_align(), but sets a path and doesn't draw anything.
        Call stroke() or fill() yourself.
        """
        (dx, dy, w, h, ax, ay) = self.cr.text_extents(text_string)
        if centered:
            x = x - w / 2
            y = y + h / 2
        # Let text_path() do the drawing
        self.text_path(text_string, x, y)


    #def gradient_units(self, units):
    #    """Set gradient units to one of:
    #    userSpace, userSpaceOnUse, objectBoundingBox
    #    """
    #    self.insert('gradient-units %s' % units)


    #def gravity(self, direction):
    #    """Set the gravity direction to one of:
    #    
    #        NorthWest, North, NorthEast,
    #        West, Center, East,
    #        SouthWest, South, SouthEast
    #        
    #    Note: gravity is known to affect text and images; it does not affect
    #    points, lines, circles, rectangles, roundrectangles, polylines or
    #    polygons.
    #    """
    #    self.insert('gravity %s' % direction)


    #def matte(self, (x, y), method='floodfill'):
    #    # method may be: point, replace, floodfill, filltoborder, reset
    #    # (What do x, y mean?)
    #    self.insert('matte %s,%s %s' % (x, y, method))


    #def offset(self, offset):
    #    self.insert('offset %s' % offset)


    #def path(self, path_data):
    #    """Draw a path. path_data is a list of instructions and coordinates,
    #    e.g. ['M', (100,100), 'L', (200,200), ...]
    #    Instructions are M (moveto), L (lineto), A (arc), Z (close path)
    #    For more on using paths (and a complete list of commands), see:
    #    http://www.cit.gu.edu.au/~anthony/graphics/imagick6/draw/#paths
    #    http://www.w3.org/TR/SVG/paths.html#PathDataGeneralInformation
    #    """
    #    command = 'path'
    #    for token in path_data:
    #        if isinstance(token, tuple):
    #            command += ' %s,%s' % token
    #        else:
    #            command += ' %s' % token
    #    self.insert(command)


    #def point(self, (x, y), size):
    #    """Draw a point at position (x, y)"""
    #    self.insert('point %s,%s' % (x, y))



### --------------------------------------------------------------------
### Exported functions
### --------------------------------------------------------------------


# Cached masks, rendered only once.
interlace_fields = None


def interlace_drawings(draw1, draw2):
    """Merge two drawings into one, using interlacing fields.

    This method interlaces with the BOTTOM-FIRST field order.
    """
    global interlace_fields
    
    dr = Drawing(draw1.w, draw1.h)

    # create masks (only once)
    if (not interlace_fields):
        fields = [0, 1]
        for f in range(0,2):
            img = cairo.ImageSurface(cairo.FORMAT_ARGB32, draw1.w, draw1.h)
            cr = cairo.Context(img)
            cr.set_antialias(cairo.ANTIALIAS_NONE)
            cr.set_source_rgba(0.5, 0.5, 0.5, 1)
            cr.set_line_width(1)

            for x in range(0, draw1.h / 2):
                # x*2 + f = top field first.., each two lines..
                # -1 à draw1.w +1 -> pour être sur de couvrir aussi les bouts.
                cr.move_to(-1, x*2 + f)
                cr.line_to(draw1.w + 1, x*2 + f)
                cr.stroke()

            fields[f] = img
        interlace_fields = fields
    else:
        fields = interlace_fields

    # For bottom first, use fields[0] first, and fields[1] after.
    # For top-first, use draw1 with fields[1] and draw2 with fields[0]
    # paint first image
    dr.image_surface(0,0,dr.w, dr.h, draw1.surface, fields[0])
    # paint second image
    dr.image_surface(0,0,dr.w, dr.h, draw2.surface, fields[1])

    return dr



def render(drawing, context, width, height):
    """Render a Drawing to the given cairo context at the given size.
    """
    # Scale from the original size
    scale_x = float(width) / drawing.size[0]
    scale_y = float(height) / drawing.size[1]
    drawing.scale(scale_x, scale_y)
    # Sneaky bit--scaling needs to happen before everything else
    drawing.steps.insert(0, drawing.steps.pop(-1))
    
    # Execute all draw commands
    for step in drawing.steps:
        log.debug("Drawing step: %s" % step)
        step.function(context)

    # Remove scaling function
    drawing.steps.pop(0)


def display(drawing, width=None, height=None, fix_aspect=False):
    """Render and display the given Drawing at the given size.
    
        drawing:    A Drawing object to display
        width:      Pixel width of displayed image,
                    or 0 to use the given Drawing's size
        height:     Pixel height of displayed image,
                    or 0 to use the given Drawing's size
        fix_aspect: True to adjust height to make the image
                    appear in the correct aspect ratio
    """
    if not width:
        width = drawing.size[0]
    if not height:
        height = drawing.size[1]

    # Adjust height to fix aspect ratio if desired
    if fix_aspect:
        height = width / drawing.aspect
    # Render and display a .png
    png_file = '/tmp/drawing.png'
    save_png(drawing, png_file, width, height)
    print "Displaying", png_file, "at %sx%s" % (width, height)
    print "(press 'q' in image window to continue)"
    print commands.getoutput('display %s' % png_file)


def write_ppm(drawing, pipe, width, height, workdir=None):
    """Write image as a PPM file to a file-object

      workdir - Unused in this context, just to have parallel parameters
                with write_png().

    Useful to pipe directly in ppmtoy4m and to pipe directly to mpeg2enc.
    """
    # Timing
    start = time.time()
    if (width, height) == drawing.size:
        #print "Not re-rendering"
        surface = drawing.surface
    else:
        surface = get_surface(width, height, 'image')
        context = cairo.Context(surface)
        render(drawing, context, width, height)
        
    buf = surface.get_data()
    # Assumes surface is cairo.FORMAT_ARGB32
    im = Image.frombuffer('RGBA', (surface.get_width(),  surface.get_height()),
                          buf)
    im = im.transpose(Image.FLIP_TOP_BOTTOM)
    im.save(pipe, 'ppm')
    
    print "write_ppm took %s seconds" % (time.time() - start)
    
def write_png(drawing, pipe, width, height, workdir):
    """Write image as a PPM file to a file-object

    Useful to pipe directly in pngtopnm -mix -background=rgb:00/00/00 | ppmtoy4m | mpeg2enc.
    """
    # Timing
    start = time.time()
    if (width, height) == drawing.size:
        #print "Not re-rendering"
        surface = drawing.surface
    else:
        surface = get_surface(width, height, 'image')
        context = cairo.Context(surface)
        render(drawing, context, width, height)

    surface.write_to_png('%s/tmp.png' % workdir)
    im = Image.open('%s/tmp.png' % workdir)
    im.load()
    im.save(pipe, 'ppm')
    
    print "write_png took %s seconds" % (time.time() - start)




def save_png(drawing, filename, width, height):
    """Saves a drawing to PNG, keeping alpha channel intact.

    This is the quickest, since Cairo itself exports directly to .png"""
    # Timing
    start = time.time()
    if (width, height) == drawing.size:
        print "Not re-rendering"
        surface = drawing.surface
    else:
        surface = get_surface(width, height, 'image')
        context = cairo.Context(surface)
        render(drawing, context, width, height)
    surface.write_to_png(filename)
    print "save_png took %s seconds" % (time.time() - start)


def save_jpg(drawing, filename, width, height):
    """Saves a drawing to JPG, losing alpha channel information.
    """
    f = open('/tmp/export.png', 'wb+')
    save_png(drawing, f, width, height)
    f.seek(0)
    im = Image.open(f)
    im.save(filename)
    del(im)
    f.close()


def save_image(drawing, img_filename, width, height):
    """Render drawing to a .jpg, .png or other image."""
    log.info("Saving Drawing to %s" % img_filename)
    if img_filename.endswith('.png'):
        save_png(drawing, img_filename, width, height)
    elif img_filename.endswith('.jpg'):
        save_jpg(drawing, img_filename, width, height)
    else:
        log.debug("Creating temporary .png")
        temp_png = '/tmp/%s.png' % img_filename
        save_png(drawing, temp_png, width, height)
        log.debug("Converting temporary png to %s" % img_filename)
        cmd = "convert -size %sx%s " % drawing.size
        cmd += "%s %s" % (temp_png, img_filename)
        print commands.getoutput(cmd)


def save_svg(drawing, filename, width, height):
    """Render drawing to an SVG file."""
    surface = get_surface(width, height, 'svg', filename)
    context = cairo.Context(surface)
    render(drawing, context, width, height)
    context.show_page()
    surface.finish()
    print "Saved SVG to", filename


def save_pdf(drawing, filename, width, height):
    """Render drawing to a PDF file."""
    surface = get_surface(width, height, 'pdf', filename)
    context = cairo.Context(surface)
    render(drawing, context, width, height)
    context.show_page()
    surface.finish()
    print "Saved PDF to", filename


def save_ps(drawing, filename, width, height):
    """Render drawing to a PostScript file."""
    surface = get_surface(width, height, 'ps', filename)
    context = cairo.Context(surface)
    render(drawing, context, width, height)
    context.show_page()
    surface.finish()
    print "Saved PostScript to", filename


### --------------------------------------------------------------------
### Demo functions
### --------------------------------------------------------------------

def draw_fontsize_demo(drawing):
    """Draw font size samples on the given drawing."""
    assert isinstance(drawing, Drawing)

    # Save context
    drawing.save()
    # TODO: Remove this scaling hack
    drawing.scale(800.0/720, 600.0/480)
    
    # Draw white text in a range of sizes
    drawing.set_source('white')
    
    for size in [12,16,20,24,28,32]:
        ypos = size * size / 5
        drawing.font('Nimbus Sans')
        drawing.font_size(size)
        drawing.text("%s pt: The quick brown fox" % size, 700, ypos, 'right')
    
    # Restore context
    drawing.restore()


def draw_font_demo(drawing):
    """Draw samples of different fonts on the given drawing."""
    assert isinstance(drawing, Drawing)

    # Save context
    drawing.save()
    # TODO: Remove this scaling hack
    drawing.scale(800.0/720, 600.0/480)

    fontsize = 24
    drawing.font_size(fontsize)
    fonts = [
        'Arial',
        'Baskerville',
        'Dragonwick',
        'Georgia',
        'Helvetica',
        'Linotext',
        'Luxi Mono',
        'Nimbus Sans',
        'Old-Town',
        'Sharktooth',
        'Tahoma',
        'Times']
    ypos = 0
    for font in fonts:
        drawing.font(font)
        # Transparent shadow
        drawing.set_source('darkblue', 0.4)
        drawing.text(font, 3, ypos+3)
        # White text
        drawing.set_source('white', 1.0)
        drawing.text(font, 0, ypos)
        ypos += fontsize

    # Restore context
    drawing.restore()


def draw_shape_demo(drawing):
    """Draw shape samples on the given drawing."""
    assert isinstance(drawing, Drawing)

    # Save context
    drawing.save()
    # TODO: Remove this scaling hack
    drawing.scale(800.0/720, 600.0/480)

    # Large orange circle with black stroke
    drawing.save()
    drawing.stroke_width(12)
    # Circle at (500, 200), with radius 200
    drawing.circle(500, 200, 200)
    drawing.stroke('black')
    drawing.fill('orange')
    drawing.restore()

    # Grey-stroked blue circles
    drawing.save()
    # TODO:
    # All circles have the same stroke width and color (and, in other cases,
    # might all have the same fill style as well). A simpler interface is
    # called for, e.g.:
    # drawing.set_style(fill='#8080FF', stroke='#777', stroke_width=2)
    # drawing.circle((65, 50), 15)
    # drawing.circle((65, 100), 10)
    # drawing.circle((65, 150), 5)
    # drawing.set_style(fill='black') # stroke stays the same as before
    # drawing.rectangle((40, 30), (50, 150))
    
    drawing.stroke_width(2)
    drawing.circle(65, 50, 15)
    drawing.fill('#8080FF')
    drawing.stroke('#777')
    
    drawing.circle(60, 100, 10)
    drawing.fill('#2020F0')
    drawing.stroke('#777')
    
    drawing.circle(55, 150, 5)
    drawing.fill('#0000A0')
    drawing.stroke('#777')
    drawing.restore()

    # Semitransparent green rectangles
    drawing.save()
    drawing.translate(50, 400)
    for scale in [0.2, 0.4, 0.7, 1.1, 1.6, 2.2, 2.9, 3.7]:
        drawing.save()
        drawing.translate(scale * 70, scale * -50)
        drawing.scale(scale, scale)
        drawing.stroke_width(scale)
        # roundrectangle broken?
        drawing.roundrectangle(-30, -30, 30, 30, 8, 8)
        #drawing.rectangle(-30, -30, 30, 30)
        drawing.fill('lightgreen', scale / 5.0)
        drawing.stroke('black')
        drawing.restore()
    drawing.restore()

    # Restore context
    drawing.restore()


def draw_stroke_demo(drawing):
    """Draw a stroke/strokewidth demo on the given drawing."""
    assert isinstance(drawing, Drawing)

    # Save context
    drawing.save()
    # TODO: Remove this scaling hack
    drawing.scale(800.0/720, 600.0/480)
    
    for width in [1, 2, 4, 6, 8, 10, 12, 14, 16]:
        drawing.stroke_width(width)
        rgb = ((255 - width * 8), (120 - width * 5), 0)
        offset = width * 10
        drawing.line(0, offset, -offset, offset)
        drawing.stroke('rgb(%s,%s,%s)' % rgb)

    # Restore context
    drawing.restore()

log.level = 'info'

if __name__ == '__main__':
    mytime = time.time() # Benchmark

    drawing = Drawing(800, 600)

    # Start a new
    drawing.save()

    # Add a background fill
    #->Background
    drawing.set_source('darkgray')
    drawing.rectangle(0, 0, 800, 600)
    drawing.fill()

    # Shape demo
    drawing.save()
    draw_shape_demo(drawing)
    drawing.restore()

    # Font size demo
    drawing.save()
    draw_fontsize_demo(drawing)
    drawing.restore()

    # Stroke demo
    drawing.save()
    drawing.translate(680, 240)
    draw_stroke_demo(drawing)
    drawing.restore()

    # Font face demo
    drawing.save()
    drawing.translate(60, 160)
    draw_font_demo(drawing)
    drawing.restore()

    # Close out the Cairo rendering...
    drawing.restore()

    print "Took %f seconds" % (time.time() - mytime)

    # Render and display the Drawing at several different sizes
    resolutions = [(352, 240), (352, 480), (720, 480), (800, 600)]
    #resolutions = [(720, 480)]
    for w, h in resolutions:
        display(drawing, w, h)

    #save_svg(drawing, "/tmp/drawing.svg", 400, 300)
    #save_pdf(drawing, "/tmp/drawing.pdf", 400, 300)
    #save_ps(drawing, "/tmp/drawing.ps", 400, 300)



#! /usr/bin/env python
# tooltip.py

# Stolen from: http://tkinter.unpythonic.net/wiki/ToolTip
"""
Michael Lange <klappnase at 8ung dot at>

The ToolTip class provides a flexible tooltip widget for Tkinter; it is
based on IDLE's ToolTip module which unfortunately seems to be broken
(at least the version I saw).

INITIALIZATION OPTIONS:

    anchor:         where the text should be positioned inside the widget, 
                    must be on of "n", "s", "e", "w", "nw" and so on; default
                    is "center"
    bd:             borderwidth of the widget; default is 1 (NOTE: don't
                    use "borderwidth" here)
    bg:             background color to use for the widget; default is
                    "lightyellow" (NOTE: don't use "background")
    delay:          time in ms that it takes for the widget to appear on the
                    screen when the mouse pointer has entered the parent
                    widget; default is 1500
    fg:             foreground (i.e. text) color to use; default is "black"
                    (NOTE: don't use "foreground")
    follow_mouse:   if set to 1 the tooltip will follow the mouse pointer
                    instead of being displayed outside of the parent widget;
                    this may be useful if you want to use tooltips for
                    large widgets like listboxes or canvases; default is 0
    font:           font to use for the widget; default is system specific
    justify:        how multiple lines of text will be aligned, must be
                    "left", "right" or "center"; default is "left"
    padx:           extra space added to the left and right within the widget; 
                    default is 4
    pady:           extra space above and below the text; default is 2
    relief:         one of "flat", "ridge", "groove", "raised", "sunken" or
                    "solid"; default is "solid"
    state:          must be "normal" or "disabled"; if set to "disabled" the
                    tooltip will not appear; default is "normal"
    text:           the text that is displayed inside the widget
    textvariable:   if set to an instance of Tkinter.StringVar() the variable's
                    value will be used as text for the widget
    width:          width of the widget; the default is 0, which means that
                    "wraplength" will be used to limit the widgets width
    wraplength:     limits the number of characters in each line;
                    default is 150

WIDGET METHODS:
    configure(**opts):  change one or more of the widget's options as described
                        above; the changes will take effect the next time the
                        tooltip shows up; NOTE: follow_mouse cannot be changed
                        after widget initialization
    
Other widget methods that might be useful if you want to subclass ToolTip:
    enter():            callback when the mouse pointer enters the parent widget
    leave():            called when the mouse pointer leaves the parent widget
    motion():           is called when the mouse pointer moves inside the
                        parent widget if follow_mouse is set to 1 and the
                        tooltip has shown up to continually update the
                        coordinates of the tooltip window
    coords():           calculates the screen coordinates of the tooltip window
    create_contents():  creates the contents of the tooltip window (by default
                        a Tkinter.Label)
"""

__all__ = ['ToolTip']

# Ideas gleaned from PySol

import Tkinter

class ToolTip:
    _defaults = {
        'anchor': 'center',
        'bd': 1,
        'bg': 'lightyellow',
        'fg': 'black',
        'follow_mouse': 0,
        'font': None,
        'justify': 'left',
        'padx': 4,
        'pady': 2,
        'relief': 'solid',
        'state': 'normal',
        'textvariable': None,
        'width': 0,
        'wraplength': 150
    }
    def __init__(self, master, text='Your text here', delay=1500, **opts):
        self.master = master
        self._opts = self._defaults.copy()
        self._opts.update({'delay': delay, 'text': text})
        self.configure(**opts)
        self._tipwindow = None
        self._id = None
        self._id1 = self.master.bind("<Enter>", self.enter, '+')
        self._id2 = self.master.bind("<Leave>", self.leave, '+')
        self._id3 = self.master.bind("<ButtonPress>", self.leave, '+')
        self._follow_mouse = 0
        if self._opts['follow_mouse']:
            self._id4 = self.master.bind("<Motion>", self.motion, '+')
            self._follow_mouse = 1

    def configure(self, **opts):
        for key in opts:
            if self._opts.has_key(key):
                self._opts[key] = opts[key]
            else:
                KeyError = 'KeyError: Unknown option: "%s"' %key
                raise KeyError

    ### Event handlers

    def enter(self, event=None):
        self._schedule()

    def leave(self, event=None):
        self._unschedule()
        self._hide()

    def motion(self, event=None):
        if self._tipwindow and self._follow_mouse:
            x, y = self.coords()
            self._tipwindow.wm_geometry("+%d+%d" % (x, y))

    ### Helpers

    def _schedule(self):
        self._unschedule()
        if self._opts['state'] == 'disabled':
            return
        self._id = self.master.after(self._opts['delay'], self._show)

    def _unschedule(self):
        id = self._id
        self._id = None
        if id:
            self.master.after_cancel(id)

    def _show(self):
        if self._opts['state'] == 'disabled':
            self._unschedule()
            return
        if not self._tipwindow:
            self._tipwindow = tw = Tkinter.Toplevel(self.master)
            # hide the window until we know the geometry
            tw.withdraw()
            tw.wm_overrideredirect(1)

            if tw.tk.call("tk", "windowingsystem") == 'aqua':
                tw.tk.call("::tk::unsupported::MacWindowStyle",
                           "style", tw._w, "help", "none")

            self.create_contents()
            tw.update_idletasks()
            x, y = self.coords()
            tw.wm_geometry("+%d+%d" % (x, y))
            tw.deiconify()

    def _hide(self):
        tw = self._tipwindow
        self._tipwindow = None
        if tw:
            tw.destroy()

    ##these methods might be overridden in derived classes:##

    def coords(self):
        # The tip window must be completely outside the master widget;
        # otherwise when the mouse enters the tip window we get
        # a leave event and it disappears, and then we get an enter
        # event and it reappears, and so on forever :-(
        # or we take care that the mouse pointer is always outside the tipwindow :-)
        tw = self._tipwindow
        twx, twy = tw.winfo_reqwidth(), tw.winfo_reqheight()
        w, h = tw.winfo_screenwidth(), tw.winfo_screenheight()
        # calculate the y coordinate:
        if self._follow_mouse:
            y = tw.winfo_pointery() + 20
            # make sure the tipwindow is never outside the screen:
            if y + twy > h:
                y = y - twy - 30
        else:
            y = self.master.winfo_rooty() + self.master.winfo_height() + 3
            if y + twy > h:
                y = self.master.winfo_rooty() - twy - 3
        # we can use the same x coord in both cases:
        x = tw.winfo_pointerx() - twx / 2
        if x < 0:
            x = 0
        elif x + twx > w:
            x = w - twx
        return x, y

    def create_contents(self):
        opts = self._opts.copy()
        for opt in ('delay', 'follow_mouse', 'state'):
            del opts[opt]
        label = Tkinter.Label(self._tipwindow, **opts)
        label.pack()

##---------demo code-----------------------------------##

def demo():
    root = Tkinter.Tk(className='ToolTip-demo')
    l = Tkinter.Listbox(root)
    l.insert('end', "I'm a listbox")
    l.pack(side='top')
    t1 = ToolTip(l, follow_mouse=1, text="I'm a tooltip with follow_mouse set to 1, so I won't be placed outside my parent")
    b = Tkinter.Button(root, text='Quit', command=root.quit)
    b.pack(side='bottom')
    t2 = ToolTip(b, text='Enough of this')
    root.mainloop()

if __name__ == '__main__':
    demo()


#! /usr/bin/env python
# media.py

"""This module provides a multimedia file class (MediaFile), for storing
a profile of attributes including resolution, audio and video codecs and
bitrates.

Two functions are also provided:

    load_media(filename)
        Return a MediaFile filled with attributes from the given file
    standard_media(format, tvsys)
        Return a MediaFile profile matching the given format and TV system

These can be used for getting source and target MediaFiles for encoding via
one of the backends in libtovid.transcode.encode. For example:

    >>> dvd = standard_media('dvd', 'ntsc')
    >>> print dvd
    Filename:
    Format: DVD
    TVsys: NTSC
    Length: 0 seconds
    Video: 720x480 29.97fps 4:3 mpeg2 6000kbps
    Audio: ac3 224kbps 48000hz 2-channel

"""

__all__ = [\
    'MediaFile',
    'load_media',
    'standard_media',
    'correct_aspect']

# From standard library
import os
import sys
import copy
import commands
# From libtovid
from libtovid import log
from libtovid.cli import Command
from libtovid import standard
from libtovid.utils import ratio_to_float

class MediaFile:
    """A multimedia video file, and its vital statistics.
    """
    def __init__(self, filename='', format='dvd', tvsys='ntsc'):
        log.debug("MediaFile(%s, %s, %s)" % (filename, format, tvsys))
        # TODO: Set attributes to match given format and tvsys
        self.filename = filename
        self.format = format
        self.tvsys = tvsys
        self.length = 0
        # TODO: Support multiple audio and video tracks
        self.has_audio = False
        self.has_video = False
        # Audio attributes
        self.acodec = 'ac3'
        self.abitrate = 224
        self.channels = 2
        self.samprate = 48000
        # Video attributes
        self.vcodec = 'mpeg2'
        # TODO: Better names for scale & expand, but not so long as
        # 'inner_resolution' and 'outer_resolution', or
        # 'picture_res' and 'frame_res'. Suggestions?
        self.scale = (720, 480)
        self.expand = (720, 480)
        self.vbitrate = 9800
        self.fps = 29.97
        self.aspect = '4:3'
        self.widescreen = False

    def __str__(self):
        """Return a string representation of the MediaFile suitable for
        printing.
        """
        width, height = self.expand or self.scale
        # List of lines of output
        lines = [\
            'Filename: %s' % self.filename,
            'Format: %s' % self.format.upper(),
            'TVsys: %s' % self.tvsys.upper(),
            'Length: %s seconds' % self.length,
            'Video: %sx%s %sfps %s %s %skbps' % \
            (width, height, self.fps, self.aspect, self.vcodec, self.vbitrate),
            'Audio: %s %skbps %shz %s-channel' % \
            (self.acodec, self.abitrate, self.samprate, self.channels)
            ]
        # Add newlines and return
        return '\n'.join(lines)


def load_media(filename, length_accuracy=False):
    """Return a MediaFile filled with attributes read from a file.
    
        filename:  Name of a multimedia video file
        
    """
    log.debug("load_media(%s)" % filename)
    # TODO: Raise an exception if the file couldn't be identified
    # TODO: Infer aspect ratio
    media = MediaFile(filename)
    mp_dict = {}
    # Use mplayer
    cmd = Command('mplayer',
                  '-identify',
                  '%s' % filename,
                  '-vo', 'null',
                  '-ao', 'null',
                  '-frames', '1',
                  '-channels', '6')
    cmd.run(capture=True)
    # Look for mplayer's "ID_..." lines and include each assignment in mp_dict
    for line in cmd.get_output().splitlines():
        if line.startswith("ID_"):
            left, right = line.split('=')
            mp_dict[left] = right.strip()
            
    # Check for existence of streams
    if 'ID_VIDEO_ID' in mp_dict:
        media.has_video = True
    else:
        media.has_video = False
    if 'ID_AUDIO_ID' in mp_dict:
        media.has_audio = True
    else:
        media.has_audio = False

    # Parse the dictionary and set appropriate values
    for left, right in mp_dict.iteritems():
        if left == "ID_VIDEO_WIDTH":
            media.scale = (int(right), media.scale[1])
        elif left == "ID_VIDEO_HEIGHT":
            media.scale = (media.scale[0], int(right))
        elif left == "ID_VIDEO_FPS":
            media.fps = float(right)
        elif left == "ID_VIDEO_FORMAT":
            media.vcodec = right
        elif left == "ID_VIDEO_BITRATE":
            media.vbitrate = int(right) / 1000
        elif left == "ID_AUDIO_CODEC":
            media.acodec = right
        elif left == "ID_AUDIO_FORMAT":
            audio_format = right
        elif left == "ID_AUDIO_BITRATE":
            media.abitrate = int(right) / 1000
        elif left == "ID_AUDIO_RATE":
            media.samprate = int(right)
        elif left == "ID_AUDIO_NCH":
            media.channels = right
        elif left == 'ID_LENGTH':
            media.length = float(right)
    media.expand = media.scale

    # Length accuracy
    if (length_accuracy):
        media.length = float(commands.getoutput("mencoder '%s' -quiet -ovc copy -oac pcm -o /dev/null 2>/dev/null | awk '/Video stream/ {print $10}'" % filename))

    
    # Fix mplayer's audio codec naming for ac3 and mp2
    if media.acodec == "8192":
        media.acodec = "ac3"
    elif media.acodec == "80":
        media.acodec = "mp2"
    # Fix mplayer's video codec naming for mpeg1 and mpeg2
    if media.vcodec == "0x10000001":
        media.vcodec = "mpeg1"
    elif media.vcodec == "0x10000002":
        media.vcodec = "mpeg2"
    return media


def standard_media(format, tvsys):
    """Return a MediaFile compliant with a standard format and TV system.
    
        format:  Standard format ('vcd', 'svcd', or 'dvd')
        tvsys:   TV system ('pal' or 'ntsc')

    """
    log.debug("standard_media(%s, %s)" % (format, tvsys))
    media = MediaFile('', format, tvsys)
    # Set valid video attributes
    media.vcodec = standard.vcodec(format)
    media.scale = standard.resolution(format, tvsys)
    media.expand = media.scale
    media.fps = standard.fps(tvsys)
    # Set valid audio attributes
    media.acodec = standard.acodec(format)
    media.samprate = standard.samprate(format)
    # TODO: How to handle default bitrates? These functions return a range.
    #media.vbitrate = standard.vbitrate(format)[1]
    #media.abitrate = standard.abitrate(format)[1]
    media.abitrate = 224
    if format == 'vcd':
        media.vbitrate = 1150
    elif format == 'svcd':
        media.vbitrate = 2600
    else:
        media.vbitrate = 6000
    return media

def correct_aspect(source, target, aspect='auto'):
    """Calculate the necessary scaling to fit source into target at a given
    aspect ratio, without distorting the picture.

        source:  Input MediaFile
        target:  Output MediaFile
        aspect:  Aspect ratio to assume for input file (e.g., '4:3', '16:9')
                 or 'auto' to use autodetection

    Return a new target MediaFile with correct scaling, using letterboxing
    if appropriate, and anamorphic widescreen if available.
    """
    assert isinstance(source, MediaFile)
    assert isinstance(target, MediaFile)

    # Base new target on existing one
    target = copy.copy(target)
    
    # Convert aspect (ratio) to a floating-point value
    if aspect == 'auto':
        src_aspect = ratio_to_float(source.aspect)
    else:
        src_aspect = ratio_to_float(aspect)
    
    # Use anamorphic widescreen for any video ~16:9 or wider
    # (Only DVD supports this)
    if src_aspect >= 1.7 and target.format == 'dvd':
        target_aspect = 16.0/9.0
        widescreen = True
    else:
        target_aspect = 4.0/3.0
        widescreen = False

    width, height = target.scale
    # If aspect matches target, no letterboxing is necessary
    tolerance = 0.05
    if abs(src_aspect - target_aspect) < tolerance:
        scale = (width, height)
    # If aspect is wider than target, letterbox vertically
    elif src_aspect > target_aspect:
        scale = (width, int(height * target_aspect / src_aspect))
    # Otherwise (rare), letterbox horizontally
    else:
        scale = (int(width * src_aspect / target_aspect), height)
    expand = (width, height)

    # If input file is already the correct size, don't scale
    if scale == source.scale:
        scale = False

    # Final scaling/expansion for correct aspect ratio display
    target.scale = scale
    target.expand = expand
    target.widescreen = widescreen
    return target


# Self-test; executed when this file is run standalone
if __name__ == '__main__':
    # If no arguments were provided, print usage notes
    if len(sys.argv) == 1:
        print "Usage: media.py FILE"
    else:
        print "Creating a MediaFile object from file: %s" % sys.argv[1]
        infile = load_media(sys.argv[1])
        print "------------------------------"
        print infile
        print "------------------------------"





#! /usr/bin/env python
# __init__.py (libtovid)

"""Provides a Python interface to the functionality of tovid.
"""

## \mainpage How to use libtovid
# 
# \section intro Introduction
# 
# libtovid is a Python module oriented towards creating video discs. It has
# the following submodules:
#
# - transcode
#   - rip.py
#   - encode.py
#   - subtitles.py
# - layout
# - render
#   - drawing.py
#   - animation.py
#   - layer.py
#   - effect.py
#   - flipbook.py
# - Layers in layer.py, to put something in your canvas
# - Effects in effect.py, for several effects, which you can extend
# - Flipbook for animation
# - Drawing module in drawing.py for drawing primitives
# - Tween and Keyframe in animation.py, for key-framing
#
#
# <i>You'll find this page in %libtovid/__init__.py</i>


__all__ = [\
    # Subdirectories
    'author',
    'gui',
    'metagui',
    'render',
    'template',
    'test',
    'transcode',
    # .py files
    'cli',
    'deps',
    'layout'
    'media',
    'odict',
    'opts',
    'output',
    'playtime',
    'spumux',
    'standard',
    'stats',
    'utils',
    'xml']

import os
import sys
from ConfigParser import ConfigParser


# Configuration file reader/writer
class Config (ConfigParser):
    """Interface for reading/writing tovid configuration file. Just a wrapper
    around the standard library ConfigParser. Example usage:
    
        config = libtovid.Config()
        config.get('DEFAULT', 'work_dir')
        config.set('tovid', 'method', 'ffmpeg')
        config.save()
        
    See the ConfigParser documentation for details.
    """
    # Dictionary of suite-wide configuration defaults
    DEFAULTS = {
        'work_dir': '/tmp',
        'output_dir': '/tmp'}

    def __init__(self):
        """Load configuration from ~/.tovid/config."""
        ConfigParser.__init__(self, self.DEFAULTS)
        self.filename = os.path.expanduser('~/.tovid/config')
        self.read(self.filename)

    def save(self):
        """Save the configuration to the current filename."""
        outfile = open(self.filename, 'w')
        outfile.write('# tovid configuration file\n\n')
        self.write(outfile)
        outfile.close()


# Logging class
class Log:
    # Increasing levels of severity
    _levels = {
        'debug': 1,
        'info': 2,
        'warning': 3,
        'error': 4,
        'critical': 5}
    
    def __init__(self, level='info'):
        """Create a logger with the given severity level."""
        self.level = level
    
    def message(self, level, text):
        """Print a message if it's at the current severity level or higher."""
        if Log._levels[level] >= Log._levels[self.level]:
            print text

    def debug(self, text):
        """Log a debugging message."""
        self.message('debug', text)

    def info(self, text):
        """Log an informational message."""
        self.message('info', text)

    def warning(self, text):
        """Log a warning message."""
        self.message('warning', "WARNING: " + text)

    def error(self, text):
        """Log an error message."""
        self.message('error', "ERROR: " + text)

    def critical(self, text):
        """Log a critical error message."""
        self.message('critical', "CRITICAL: " + text)

# Global logger
log = Log()



#! /usr/bin/env python
# stats.py

"""Classes and functions for dealing with video statistics.

Future interface ideas:

* Display certain records, or a range of records (records 1-10; last 15 records;
  records with a certain field (or any field) matching given text/regexp
* Display formatted output of all records, or a selection of records,
  with control over which fields are displayed

"""

import csv
import sys
from copy import copy

# Order of fields in stats.tovid
FIELDS = [\
    'tovid_version',
    'final_name',
    'length',          # a.k.a. CUR_LENGTH
    'format',          # RES
    'tvsys',
    'final_size',
    'tgt_bitrate',
    'avg_bitrate',
    'peak_bitrate',
    'gop_minsize',
    'gop_maxsize',
    'encoding_time',   # SCRIPT_TOT_TIME
    'cpu_model',
    'cpu_speed',
    'in_vcodec',
    'in_acodec',
    'encoding_mode',
    'in_md5',
    'in_width',
    'in_height',
    'quant',
    'kbpm',
    'enc_time_ratio',
    'backend'
    ]

# Integer numeric fields
int_fields = [\
    'length',
    'avg_bitrate',
    'tgt_bitrate',
    'peak_bitrate',
    'encoding_time',
    'final_size',
    'cpu_speed',
    'in_width',
    'in_height',
    'quant',
    'kbpm'
    ]

class Statlist:
    """A list of statistics that may be queried with a simple database-like
    interface."""
    def __init__(self, records=None, filename=''):
        """Create a Statlist, using the given list of records, or by reading
        from the given filename (CSV text).
        """
        # Use provided records, if any
        if records:
            self.records = copy(records)
        # Otherwise, read from any provided filename
        else:
            self.records = []
            if filename is not '':
                self.read_csv(filename)

    def read_csv(self, filename):
        """Import stats from a CSV (comma-delimited quoted text) file."""
        self.records = []
        statfile = open(filename, 'r')
        csv_reader = csv.DictReader(statfile, FIELDS, skipinitialspace=True)
        for line in csv_reader:
            # Convert some string and numeric fields
            line['format'] = str.lower("%s" % line['format'])
            line['tvsys'] = str.lower("%s" % line['tvsys'])
            for field in int_fields:
                try:
                    num = int("%s" % line[field])
                except ValueError, TypeError:
                    num = 0
                line[field] = num
            self.records.append(line)

        statfile.close()
        print "Read %s lines from %s" % (len(self.records), filename)

    def unique(self, field):
        """Return a list of unique values of the given field."""
        unique_values = []
        for record in self.records:
            if record[field] not in unique_values:
                unique_values.append(record[field])
        return unique_values

    def count_unique(self, field):
        """Count the occurrences of each unique value of the given field.
        Return a dictionary of unique values and the number of occurrences
        of each."""
        counts = {}
        # Go through all records and total up occurrences of each value
        for record in self.records:
            value = record[field]
            if value is None or value == '':
                pass
            elif value not in counts:
                counts[value] = 1
            else:
                counts[value] += 1
        return counts
        
    def average(self, attribute):
        """Calculate the average value for a given numeric VidStat attribute.
        For example, average('bitrate') returns the average overall bitrate of
        all videos in the list."""
        # Error if attribute is non-numeric
        if attribute not in int_fields:
            print "Can't average %s: not defined as a numeric field"
            sys.exit()
            
        values = []
        for record in self.records:
            # Only append non-zero values
            if record[attribute] != 0:
                values.append(record[attribute])
        if len(values) > 0:
            return float(sum(values) / len(values))
        else:
            return 0

    def average_by(self, attribute, by_attribute):
        """Return a dictionary of averages of an attribute, indexed by another
        attribute. For example, average_by('bitrate', 'format') returns average
        bitrates for each format."""
        # Error if attribute is non-numeric
        if attribute not in int_fields:
            print "Can't average %s: not defined as a numeric field"
            sys.exit()

        values_by = self.list_by(attribute, by_attribute)
        # Calculate averages for each value-list
        averages = {}
        if len(values_by) > 0:
            for key, samples in values_by.iteritems():
                try:
                    averages[key] = float(sum(samples) / len(samples))
                except ZeroDivisionError:
                    averages[key] = 0.0
        return averages

    def list_by(self, attribute, by_attribute, sort_lists=False):
        """Return a dictionary of lists of values of the given attribute,
        indexed by another attribute. If sort_lists is True, sort all lists."""
        # Create a dictionary of value-lists, indexed by by_attribute
        values_by = {}        
        for record in self.records:
            byval = record[by_attribute]
            if not values_by.has_key(byval):
                values_by[byval] = []
            # Only include non-zero values
            if record[attribute] != 0:
                values_by[byval].append(record[attribute])
        if sort_lists:
            for index in values_by.iterkeys():
                values_by[index].sort()
        return values_by
    
    def get_matching(self, attribute, value):
        """Return a list of records where the given attribute equals the given
        value. Records are field-indexed dictionaries of values.
        """
        # List of matching records
        matches = []
        for record in self.records:
            if record[attribute] == value:
                matches.append(record)
        return matches
    
    def length(self, field):
        """Return the length of the longest record in the given field, or the
        width of the field name itself, whichever is greater."""
        longest = len(field)
        for record in self.records:
            cur_len = len(str(record[field]))
            if cur_len > longest:
                longest = cur_len
        return longest
    
    def show(self, show_records='all', show_fields=FIELDS):
        """Print records matching given criteria, showing only the given fields.
        
            show_records:   Number of record to show, or range of numbers
            show_fields:    List of fields, by name as shown in FIELDS
        """
        # Remember field sizes (character widths)
        size = {}
        # Create field headings
        heading = ''
        for field in show_fields:
            if field in FIELDS:
                size[field] = self.length(field)
                heading += "%s " % str.ljust(field, size[field])
            else:
                print "Error: field '%s' does not exist" % field
                sys.exit()
        print heading
        # Print fields from matching records
        for record in self.records:
            # TODO: support show_records
            line = ''
            for field in show_fields:
                line += "%s " % str.ljust(str(record[field]), size[field])
            print line
        # Print heading at end too
        print heading
 


#! /usr/bin/env python
# layer.py

"""This module provides a Layer class and several derivatives. A Layer
is a graphical overlay that may be composited onto an image canvas.

Run this script standalone for a demonstration:

    $ python libtovid/layer.py

Layer subclasses may combine graphical elements, including other Layers,
into a single interface for drawing and customizing those elements. Layers may
exhibit animation, through the use of keyframed drawing commands, or through
use of the Effect class (and its subclasses, as defined in libtovid/effect.py).

Each Layer subclass provides (at least) an __init__ function, and a draw
function. For more on how to use Layers, see the Layer class definition and
template example below.
"""

__all__ = [\
    'Layer',
    'Background',
    'Text',
    'ShadedText',
    'TextBox',
    'Label',
    'VideoClip',
    'Image',
    'Thumb',
    'ThumbGrid',
    'SafeArea',
    'Scatterplot',
    'InterpolationGraph',
    'ColorBars']

import os
import sys
import math
import commands
from libtovid.utils import get_file_type
from libtovid.render.drawing import Drawing, save_image
from libtovid.render.effect import Effect
from libtovid.render.animation import Keyframe, Tween
from libtovid.media import MediaFile
from libtovid.transcode import rip
from libtovid import log

log.level = 'info'

class Layer:
    """A visual element, or a composition of visual elements. Conceptually
    similar to a layer in the GIMP or Photoshop, with support for animation
    effects and sub-Layers.
    """
    def __init__(self):
        """Initialize the layer. Extend this in derived classes to accept
        configuration settings for drawing the layer; call this function
        from any derived __init__ functions."""
        self.effects = []
        self.sublayers = []
        self._parent_flipbook = None
        self._parent_layer = None

    ###
    ### Child-parent initialization
    ###
    def _init_childs(self):
        """Give access to all descendant layers and effects to their parents.

        In layers, you can access your parent layer (if sublayed) with:
            layer._parent_layer
        and to the top Flipbook object with:
            layer._parent_flipbook
        """
        for x in range(0, len(self.effects)):
            self.effects[x]._init_parent_flipbook(self._parent_flipbook)
            self.effects[x]._init_parent_layer(self)
        for x in range(0, len(self.sublayers)):
            self.sublayers[x][0]._init_parent_flipbook(self._parent_flipbook)
            self.sublayers[x][0]._init_parent_layer(self)
            self.sublayers[x][0]._init_childs(self)

    def _init_parent_flipbook(self, flipbook):
        self._parent_flipbook = flipbook

    def _init_parent_layer(self, layer):
        self._parent_layer = layer


    ###
    ### Derived-class interface
    ###
    
    def draw(self, drawing, frame):
        """Draw the layer and all sublayers onto the given Drawing. Override
        this function in derived layers."""
        assert isinstance(drawing, Drawing)

    ###
    ### Sublayer and effect interface
    ###
    
    def add_sublayer(self, layer, position=(0, 0)):
        """Add the given Layer as a sublayer of this one, at the given position.
        Sublayers are drawn in the order they are added; each sublayer may have
        its own effects, but the parent Layer's effects apply to all sublayers.
        """
        assert isinstance(layer, Layer)
        self.sublayers.append((layer, position))

    def draw_sublayers(self, drawing, frame):
        """Draw all sublayers onto the given Drawing for the given frame."""
        assert isinstance(drawing, Drawing)
        for sublayer, position in self.sublayers:
            drawing.save()
            drawing.translate(position)
            sublayer.draw(drawing, frame)
            drawing.restore()

    def add_effect(self, effect):
        """Add the given Effect to this Layer. A Layer may have multiple effects
        applied to it; all effects apply to the current layer, and all sublayers.
        """
        assert isinstance(effect, Effect)
        self.effects.append(effect)

    def draw_with_effects(self, drawing, frame):
        """Render the entire layer, with all effects applied.
        
            drawing: A Drawing object to draw the Layer on
            frame:   The frame number that is being drawn
            
        """
        # Do preliminary effect rendering
        for effect in self.effects:
            effect.pre_draw(drawing, frame)
        # Draw the layer and sublayers
        self.draw(drawing, frame)
        # Close out effect rendering, in reverse (nested) order
        for effect in reversed(self.effects):
            effect.post_draw(drawing, frame)


# ============================================================================
# Layer template
# ============================================================================
# Copy and paste the following code to create your own Layer.
#
# Layer subclasses should define two things:
#
#     __init__():  How to initialize the layer with parameters
#     draw():   How do draw the layer on a Drawing
#
# First, declare the layer class name. Include (Layer) to indicate that your
# class is a Layer.
class MyLayer (Layer):
    """Overlapping semitransparent rectangles.
    (Modify this documentation string to describe what's in your layer)"""

    # Here's the class initialization function, __init__. Define here any
    # parameters that might be used to configure your layer's behavior or
    # appearance in some way (along with default values, if you like). Here,
    # we're allowing configuration of the fill and stroke colors, with default
    # values of 'blue' and 'black', respectively:

    def __init__(self, fill_color='blue', stroke_color='black'):
        """Create a MyLayer with the given fill and stroke colors."""
        # Initialize the base Layer class. Always do this.
        Layer.__init__(self)
        # Store the given colors, for later use
        self.fill_color = fill_color
        self.stroke_color = stroke_color

    # The draw() function is responsible for rendering the contents of the
    # layer onto a Drawing. It will use the configuration given to __init__
    # (in this case, fill and stroke colors) to render something onto a Drawing
    # associated with a particular frame number:

    def draw(self, drawing, frame):
        """Draw MyLayer contents onto the given drawing, at the given frame
        number."""

        # For safety's sake, make sure you really have a Drawing object:
        assert isinstance(drawing, Drawing)

        # Save the drawing context. This prevents any upcoming effects or
        # style changes from messing up surrounding layers in the Drawing.
        drawing.save()

        # Get a Cairo pattern source for the fill and stroke colors
        # (TODO: Make this easier, or use a simpler example)
        fc = drawing.create_source(self.fill_color, 0.6)
        sc = drawing.create_source(self.stroke_color)

        # And a stroke width of 1, say:
        drawing.stroke_width(1)

        # Now, draw something. Here, a couple of pretty semitransparent
        # rectangles, using the fill and stroke color patterns created earlier:
        drawing.rectangle(0, 0,  50, 20)
        drawing.fill(fc)
        drawing.stroke(sc)
        drawing.rectangle(15, 12,  45, 28)
        drawing.fill(fc)
        drawing.stroke(sc)

        # Be sure to restore the drawing context afterwards:
        drawing.restore()

    # That's it! Your layer is ready to use. See the Demo section at the end of
    # this file for examples on how to create and render Layers using Python.

# ============================================================================
# End of Layer template
# ============================================================================


class Background (Layer):
    """A background that fills the frame with a solid color, or an image."""
    def __init__(self, color='black', filename=''):
        Layer.__init__(self)
        self.color = color
        self.filename = filename
        
    def draw(self, drawing, frame):
        assert isinstance(drawing, Drawing)
        log.debug("Drawing Background")
        width, height = drawing.size
        drawing.save()
        # Fill drawing with an image
        if self.filename is not '':
            drawing.image(0, 0, width, height, self.filename)
        # Fill drawing with a solid color
        elif self.color:
            drawing.rectangle(0, 0, width, height)
            drawing.fill(self.color)
        drawing.restore()


### --------------------------------------------------------------------

class Image (Layer):
    """A rectangular image, scaled to the given size.

    image_source -- can be anything Drawing::image() can accept.
                    See documentation in render/drawing.py.
    """
    def __init__(self, image_source, (x, y), (width, height)):
        Layer.__init__(self)
        self.size = (width, height)
        self.image_source = image_source
        self.position = (x, y)


    def draw(self, drawing, frame=1):
        assert isinstance(drawing, Drawing)
        log.debug("Drawing Image")
        drawing.save()
        # Save the source for future calls to draw, so no further
        # processing will be necessary. And other effects can be done
        # without interferring with the original source.
        self.image_source = drawing.image(self.position[0], self.position[1],
                                          self.size[0], self.size[1],
                                          self.image_source)
        drawing.restore()


### --------------------------------------------------------------------

class VideoClip (Layer):
    """A rectangular video clip, scaled to the given size.

    TODO: num_frames should accept a range [first, end], an int (1-INT) and
    rip frames accordingly. For now, it only accepts an INT for the range 1-INT
    """
    def __init__(self, filename, (width, height), position=(0,0), num_frames=120):
        Layer.__init__(self)
        self.filename = filename
        self.mediafile = MediaFile(filename)
        self.size = (width, height)
        # List of filenames of individual frames
        self.frame_files = []
        # TODO: be able to change hardcoded default values
        self.rip_frames(1, num_frames)
        # Set (x,y) position
        assert(isinstance(position, tuple))
        self.position = position

    def rip_frames(self, start, end):
        """Rip frames from the video file, from start to end frames."""
        log.info("VideoClip: Ripping frames %s to %s" % (start, end))
        outdir = '/tmp/%s_frames' % self.filename
        self.frame_files = rip.rip_frames(self.mediafile, outdir,
                                          [start, end])

    def draw(self, drawing, frame=1):
        """Draw ripped video frames to the given drawing. For now, it's
        necessary to call rip_frames() before calling this function.
        
        Video is looped.
        """
        assert isinstance(drawing, Drawing)
        log.debug("Drawing VideoClip")
        if len(self.frame_files) == 0:
            log.error("VideoClip: need to call rip_frames() before drawing.")
            sys.exit(1)
        drawing.save()
        # Loop frames (modular arithmetic)
        if frame >= len(self.frame_files):
            frame = frame % len(self.frame_files)
        filename = self.frame_files[frame-1]
        drawing.image(self.position, self.size, filename)
        drawing.restore()


### --------------------------------------------------------------------

class Text (Layer):
    """A simple text string, with size, color and font.

    text -- UTF8 encoded string.
    """
    def __init__(self, text, position=(0, 0), color='white', fontsize=20, \
                 font='Helvetica', align='left'):
        Layer.__init__(self)
        self.text = text
        self.color = color
        self.fontsize = fontsize
        self.font = font
        self.align = align
        # Set (x,y) position
        self.position = position

    # TODO: This is gonna be pretty broken...
    def extents(self, drawing):
        """Return the extents of the text as a (x0, y0, x1, y1) tuple."""
        assert isinstance(drawing, Drawing)
        drawing.save()
        drawing.font(self.font)
        drawing.font_size(self.fontsize)
        x_bearing, y_bearing, width, height, x_adv, y_adv = \
                 drawing.text_extents(self.text)
        drawing.restore()
        # Add current layer's position to the (x,y) bearing of extents
        x0 = int(self.position[0] + x_bearing)
        y0 = int(self.position[1] + y_bearing)
        x1 = int(x0 + width)
        y1 = int(y0 + height)
        return (x0, y0, x1, y1)

    def draw(self, drawing, frame=1):
        assert isinstance(drawing, Drawing)
        log.debug("Drawing Text")
        # Drop in debugger
        drawing.save()
        drawing.font(self.font)
        drawing.font_size(self.fontsize)
        if self.color is not None:
            drawing.set_source(self.color)
        # TODO: DO something with the align !!
        drawing.text(self.text, self.position[0], self.position[1], self.align)
        drawing.restore()


### --------------------------------------------------------------------

class ShadedText (Layer):
    """A simple text string, with size, color and font.

    text -- UTF8 encoded string.
    """
    def __init__(self, text, position=(0, 0), offset=(5, 5),
                 color='white', shade_color='gray', fontsize=20,
                 font='Nimbus Sans', align='left'):
        Layer.__init__(self)
        shade_position = (position[0] + offset[0],
                          position[1] + offset[1])
        self.under = Text(text, shade_position, shade_color,
                          fontsize, font, align)
        self.over = Text(text, position, color, fontsize, font, align)

    def draw(self, drawing, frame=1):
        assert isinstance(drawing, Drawing)
        log.debug("Drawing Text")
        drawing.save()
        self.under.draw(drawing, frame)
        self.over.draw(drawing, frame)
        drawing.restore()


### --------------------------------------------------------------------

class Label (Text):
    """A text string with a rectangular background.

    You can access Text's extents() function from within here too."""
    def __init__(self, text, position=(0,0), color='white', bgcolor='#555',
                 fontsize=20, font='NimbusSans'):
        Text.__init__(self, text, position, color, fontsize, font)
        self.bgcolor = bgcolor
        # Set (x,y) position
        assert(isinstance(position, tuple))
        self.position = position

    def draw(self, drawing, frame=1):
        assert isinstance(drawing, Drawing)
        log.debug("Drawing Label")
        #(dx, dy, w, h, ax, ay) = self.extents(drawing)
        (x0, y0, x1, y1) = self.extents(drawing)
        # Save context
        drawing.save()

        # Calculate rectangle dimensions from text size/length
        width = x1 - x0
        height = y1 - y0
        # Padding to use around text
        pad = self.fontsize / 3
        # Calculate start and end points of background rectangle
        start = (-pad, -height - pad)
        end = (width + pad, pad)

        # Draw a stroked round rectangle
        drawing.save()
        drawing.stroke_width(1)
        drawing.roundrectangle(start[0], start[1], end[0], end[1], pad, pad)
        drawing.stroke('black')
        drawing.fill(self.bgcolor, 0.3)
        drawing.restore()

        # Call base Text class to draw the text
        Text.draw(self, drawing, frame)

        # Restore context
        drawing.restore()


### --------------------------------------------------------------------

class Thumb (Layer):
    """A thumbnail image or video."""
    def __init__(self, filename, (width, height), position=(0,0), title=''):
        Layer.__init__(self)
        self.filename = filename
        self.size = (width, height)
        self.title = title or os.path.basename(filename)
        # Set (x,y) position
        assert(isinstance(position, tuple))
        self.position = position

        # Determine whether file is a video or image, and create the
        # appropriate sublayer
        filetype = get_file_type(filename)
        if filetype == 'video':
            self.add_sublayer(VideoClip(filename, self.size, self.position))
        elif filetype == 'image':
            self.add_sublayer(Image(filename, self.size, self.position))
        self.lbl = Label(self.title, fontsize=15)
        self.add_sublayer(self.lbl, self.position)

    def draw(self, drawing, frame=1):
        assert isinstance(drawing, Drawing)
        log.debug("Drawing Thumb")
        drawing.save()
        (x0, y0, x1, y1) = self.lbl.extents(drawing)
        drawing.translate(0, x1-x0)
        self.draw_sublayers(drawing, frame)
        drawing.restore()


### --------------------------------------------------------------------

class ThumbGrid (Layer):
    """A rectangular array of thumbnail images or videos."""
    def __init__(self, files, titles=None, (width, height)=(600, 400),
                 (columns, rows)=(0, 0), aspect=(4,3)):
        """Create a grid of thumbnail images or videos from a list of files,
        fitting in a space no larger than the given size, with the given number
        of columns and rows. Use 0 to auto-layout columns or rows, or both 
        (default).
        """
        assert files != []
        if titles:
            assert len(files) == len(titles)
        else:
            titles = files
        Layer.__init__(self)
        self.size = (width, height)
        # Auto-dimension (using desired rows/columns, if given)
        self.columns, self.rows = \
            self._fit_items(len(files), columns, rows, aspect)
        # Calculate thumbnail size, keeping aspect
        w = (width - self.columns * 16) / self.columns
        h = w * aspect[1] / aspect[0]
        thumbsize = (w, h)

        # Calculate thumbnail positions
        positions = []
        for row in range(self.rows):
            for column in range(self.columns):
                x = column * (width / self.columns)
                y = row * (height / self.rows)
                positions.append((x, y))

        # Add Thumb sublayers
        for file, title, position in zip(files, titles, positions):
            title = os.path.basename(file)
            self.add_sublayer(Thumb(file, thumbsize, (0, 0), title), position)

    def _fit_items(self, num_items, columns, rows, aspect=(4, 3)):
        # Both fixed, nothing to calculate
        if columns > 0 and rows > 0:
            # Make sure num_items will fit (columns, rows)
            if num_items < columns * rows:
                return (columns, rows)
            # Not enough room; auto-dimension both
            else:
                log.warning("ThumbGrid: Can't fit %s items" % num_items +\
                            " in (%s, %s) grid;" % (columns, rows) +\
                            " doing auto-dimensioning instead.")
                columns = rows = 0
        # Auto-dimension to fit num_items
        if columns == 0 and rows == 0:
            # TODO: Take aspect ratio into consideration to find an optimal fit
            root = int(math.floor(math.sqrt(num_items)))
            return ((1 + num_items / root), root)
        # Rows fixed; use enough columns to fit num_items
        if columns == 0 and rows > 0:
            return ((1 + num_items / rows), rows)
        # Columns fixed; use enough rows to fit num_items
        if rows == 0 and columns > 0:
            return (columns, (1 + num_items / columns))

    def draw(self, drawing, frame=1):
        assert isinstance(drawing, Drawing)
        log.debug("Drawing ThumbGrid")
        drawing.save()
        self.draw_sublayers(drawing, frame)
        drawing.restore()


### --------------------------------------------------------------------

class SafeArea (Layer):
    """Render a safe area box at a given percentage.
    """
    def __init__(self, percent, color):
        self.percent = percent
        self.color = color
        
    def draw(self, drawing, frame=1):
        assert isinstance(drawing, Drawing)
        log.debug("Drawing SafeArea")
        # Calculate rectangle dimensions
        scale = float(self.percent) / 100.0
        width, height = drawing.size
        topleft = ((1.0 - scale) * width / 2,
                  (1.0 - scale) * height / 2)
        # Save context
        drawing.save()
        drawing.translate(topleft[0], topleft[1])
        # Safe area box
        drawing.stroke_width(3)
        drawing.rectangle(0, 0,  width * scale, height * scale)
        drawing.stroke(self.color)
        # Label
        drawing.font_size(18)
        drawing.set_source(self.color)
        drawing.text(u"%s%%" % self.percent, 10, 20)
        # Restore context
        drawing.restore()


### --------------------------------------------------------------------

class Scatterplot (Layer):
    """A 2D scatterplot of data.

    Untested since MVG move.
    """
    def __init__(self, xy_dict, width=240, height=80, x_label='', y_label=''):
        """Create a scatterplot using data in xy_dict, a dictionary of
        lists of y-values, indexed by x-value."""
        self.xy_dict = xy_dict
        self.width, self.height = (width, height)
        self.x_label = x_label
        self.y_label = y_label

    def draw(self, drawing, frame):
        """Draw the scatterplot."""
        assert isinstance(drawing, Drawing)
        log.debug("Drawing Scatterplot")
        width, height = (self.width, self.height)
        x_vals = self.xy_dict.keys()
        max_y = 0
        for x in x_vals:
            largest = max(self.xy_dict[x] or [0])
            if largest > max_y:
                max_y = largest
        # For numeric x, scale by maximum x-value
        x_is_num = isinstance(x_vals[0], int) or isinstance(x_vals[0], float)
        if x_is_num:
            x_scale = float(width) / max(x_vals)
        # For string x, scale by number of x-values
        else:
            x_scale = float(width) / len(x_vals)
        # Scale y according to maximum value
        y_scale = float(height) / max_y

        # Save context
        drawing.save()
        drawing.rectangle(0, 0, width, height)
        drawing.fill('white', 0.75)
        
        # Draw axes
        #->comment("Axes of scatterplot")
        drawing.save()
        drawing.stroke_width(2)
        drawing.line(0, 0, 0, height)
        drawing.stroke('black')
        drawing.line(0, height, width, height)
        drawing.stroke('black')
        drawing.restore()

        # Axis labels
        drawing.save()
        drawing.set_source('blue')

        drawing.save()
        for i, x in enumerate(x_vals):
            drawing.save()
            if x_is_num:
                drawing.translate(x * x_scale, height + 15)
            else:
                drawing.translate(i * x_scale, height + 15)
            drawing.rotate(30)
            drawing.text(x, 0, 0)
            drawing.restore()

        drawing.font_size(20)
        drawing.text(self.x_label, width/2, height+40)
        drawing.restore()

        drawing.save()
        drawing.text(max_y, -30, 0)
        drawing.translate(-25, height/2)
        drawing.rotate(90)
        drawing.text(self.y_label, 0, 0)
        drawing.restore()

        drawing.restore()

        # Plot all y-values for each x (as small circles)
        #->comment("Scatterplot data")
        drawing.save()
        for i, x in enumerate(x_vals):
            if x_is_num:
                x_coord = x * x_scale
            else:
                x_coord = i * x_scale
            # Shift x over slightly
            x_coord += 10
            # Plot all y-values for this x
            for y in self.xy_dict[x]:
                y_coord = height - y * y_scale
                drawing.circle(x_coord, y_coord, 3)
                drawing.fill('red', 0.2)
        drawing.restore()
        
        # Restore context
        drawing.restore()


### --------------------------------------------------------------------

class InterpolationGraph (Layer):
    # TODO: Support graphing of tuple data
    """A graph of an interpolation curve, defined by a list of Keyframes and
    an interpolation method."""
    def __init__(self, keyframes, size=(240, 80), method='linear'):
        """Create an interpolation graph of the given keyframes, at the given
        size, using the given interpolation method."""
        Layer.__init__(self)
                
        self.keyframes = keyframes
        self.size = size
        self.method = method
        # Interpolate keyframes
        self.tween = Tween(keyframes, method)

    def draw(self, drawing, frame):
        """Draw the interpolation graph, including frame/value axes,
        keyframes, and the interpolation curve."""
        assert isinstance(drawing, Drawing)
        log.debug("Drawing InterpolationGraph")
        data = self.tween.data
        # Calculate maximum extents of the graph
        width, height = self.size
        x_scale = float(width) / len(data)
        y_scale = float(height) / max(data)

        #->drawing.comment("InterpolationGraph Layer")

        # Save context
        drawing.save()

        # Draw axes
        #->drawing.comment("Axes of graph")
        drawing.save()
        drawing.stroke_width(3)
        drawing.polyline([(0, 0), (0, height), (width, height)], False)
        drawing.stroke('#ccc')
        drawing.restore()

        # Create a list of (x, y) points to be graphed
        curve = []
        x = 1
        while x <= len(self.tween.data):
            # y increases downwards; subtract from height to give a standard
            # Cartesian-oriented graph (so y increases upwards)
            point = (int(x * x_scale), int(height - data[x-1] * y_scale))
            curve.append(point)
            x += 1
        drawing.save()
        # Draw the curve
        drawing.stroke_width(2)
        drawing.polyline(curve, False)
        drawing.stroke('blue')
        drawing.restore()

        # Draw Keyframes as dotted vertical lines
        drawing.save()
        # Vertical dotted lines
        drawing.set_source('red')
        drawing.stroke_width(2)
        for key in self.keyframes:
            x = int(key.frame * x_scale)
            drawing.line(x, 0,   x, height)
            drawing.stroke('red')
            
        # Draw Keyframe labels
        drawing.set_source('white')
        for key in self.keyframes:
            x = int(key.frame * x_scale)
            y = int(height - key.data * y_scale - 3)
            drawing.text(u"(%s,%s)" % (key.frame, key.data), x, y)

        drawing.restore()

        # Draw a yellow dot for current frame
        #->drawing.comment("Current frame marker")
        drawing.save()
        pos = (frame * x_scale, height - data[frame-1] * y_scale)
        drawing.circle(pos[0], pos[1], 2)
        drawing.fill('yellow')
        drawing.restore()

        # Restore context
        drawing.restore()


### --------------------------------------------------------------------

class ColorBars (Layer):
    """Standard SMPTE color bars
    (http://en.wikipedia.org/wiki/SMPTE_color_bars)
    """
    def __init__(self, size, position=(0,0)):
        """Create color bars in a region of the given size and position.
        """
        Layer.__init__(self)
        self.size = size
        # Set (x,y) position
        assert(isinstance(position, tuple))
        self.position = position

    def draw(self, drawing, frame=1):
        assert isinstance(drawing, Drawing)
        log.debug("Drawing ColorBars")
        x, y = self.position
        width, height = self.size

        drawing.save()
        # Get to the right place and size
        drawing.translate(x, y)
        drawing.scale(width, height)
        # Video-black background
        drawing.rectangle(0, 0, 1, 1)
        drawing.fill('rgb(16, 16, 16)')

        # Top 67% of picture: Color bars at 75% amplitude
        top = 0
        bottom = 2.0 / 3
        seventh = 1.0 / 7
        size = (seventh, bottom)
        bars = [(0, top, seventh, bottom, 'rgb(191, 191, 191)'),
                (seventh, top, seventh, bottom, 'rgb(191, 191, 0)'),
                (2*seventh, top, seventh, bottom, 'rgb(0, 191, 191)'),
                (3*seventh, top, seventh, bottom, 'rgb(0, 191, 0)'),
                (4*seventh, top, seventh, bottom, 'rgb(191, 0, 191)'),
                (5*seventh, top, seventh, bottom, 'rgb(191, 0, 0)'),
                (6*seventh, top, seventh, bottom, 'rgb(0, 0, 191)')]

        # Next 8% of picture: Reverse blue bars
        top = bottom
        bottom = 0.75
        height = bottom - top
        bars.extend([(0, top, seventh, height, 'rgb(0, 0, 191)'),
                (seventh, top, seventh, height, 'rgb(16, 16, 16)'),
                (2*seventh, top, seventh, height, 'rgb(191, 0, 191)'),
                (3*seventh, top, seventh, height, 'rgb(16, 16, 16)'),
                (4*seventh, top, seventh, height, 'rgb(0, 191, 191)'),
                (5*seventh, top, seventh, height, 'rgb(16, 16, 16)'),
                (6*seventh, top, seventh, height, 'rgb(191, 191, 191)')])

        # Lower 25%: Pluge signal
        top = bottom
        bottom = 1.0
        sixth = 1.0 / 6
        height = bottom - top
        bars.extend([(0, top, 1.0, height, 'rgb(16, 16, 16)'),
                (0, top, sixth, height, 'rgb(0, 29, 66)'),
                (sixth, top, sixth, height, 'rgb(255, 255, 255)'),
                (2*sixth, top, sixth, height, 'rgb(44, 0, 92)'),
                # Sub- and super- black narrow bars
                (4*sixth, top, 0.33*sixth, height, 'rgb(7, 7, 7)'),
                (4.33*sixth, top, 0.33*sixth, height,'rgb(16, 16, 16)'),
                (4.66*sixth, top, 0.33*sixth, height, 'rgb(24, 24, 24)')])

        # Draw and fill all bars
        for x, y, width, height, color in bars:
            drawing.rectangle(x, y, width, height)
            drawing.fill(color)

        drawing.restore()



if __name__ == '__main__':
    images = None
    # Get arguments, if any
    if len(sys.argv) > 1:
        # Use all args as image filenames to ThumbGrid
        images = sys.argv[1:]

    # A Drawing to render Layer demos to
    drawing = Drawing(800, 600)
    
    # Draw a background layer
    bgd = Background(color='#7080A0')
    bgd.draw(drawing, 1)

    # Draw color bars
    bars = ColorBars((320, 240), (400, 100))
    bars.draw(drawing, 1)

    # Draw a label
    drawing.save()
    drawing.translate(460, 200)
    label = Label("tovid loves Linux")
    label.draw(drawing, 1)
    drawing.restore()

    # Draw a text layer, with position.
    text = Text("Jackdaws love my big sphinx of quartz",
                (82, 62), '#bbb')
    text.draw(drawing, 1)

    # Draw a text layer
    drawing.save()
    drawing.translate(80, 60)
    text = Text("Jackdaws love my big sphinx of quartz")
    text.draw(drawing, 1)
    drawing.restore()

    # Draw a template layer (overlapping semitransparent rectangles)
    template = MyLayer('white', 'darkblue')
    # Scale and translate the layer before drawing it
    drawing.save()
    drawing.translate(50, 100)
    drawing.scale(3.0, 3.0)
    template.draw(drawing, 1)
    drawing.restore()

    # Draw a safe area test (experimental)
    safe = SafeArea(93, 'yellow')
    safe.draw(drawing, 1)

    # Draw a thumbnail grid (if images were provided)
    if images:
        drawing.save()
        drawing.translate(350, 300)
        thumbs = ThumbGrid(images, (320, 250))
        thumbs.draw(drawing, 1)
        drawing.restore()

    # Draw an interpolation graph
    drawing.save()
    drawing.translate(60, 400)
    # Some random keyframes to graph
    keys = [Keyframe(1, 25), Keyframe(10, 5), Keyframe(30, 35),
            Keyframe(40, 35), Keyframe(45, 20), Keyframe(60, 40)]
    interp = InterpolationGraph(keys, (400, 120), method="cosine")
    interp.draw(drawing, 25)
    drawing.restore()

    # Draw a scatterplot
    drawing.save()
    xy_data = {
        5: [2, 4, 6, 8],
        10: [3, 5, 7, 9],
        15: [5, 9, 13, 17],
        20: [8, 14, 20, 26]}
    drawing.translate(550, 350)
    #drawing.scale(200, 200)
    plot = Scatterplot(xy_data, 200, 200, "Spam", "Eggs")
    plot.draw(drawing, 1)
    drawing.restore()
    
    log.info("Output to /tmp/my.png")
    save_image(drawing, '/tmp/my.png', 800, 600)
